from datetime import date
import unittest
import utils
import os
import datetime
from io import StringIO

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import modelcif.reader
import modelcif.reference
import ihm
import ihm.reader


ASYM_ENTITY = """
loop_
_entity_poly_seq.entity_id
_entity_poly_seq.num
_entity_poly_seq.mon_id
_entity_poly_seq.hetero
1 1 MET .
1 2 CYS .
1 3 MET .
1 4 SER .
#
loop_
_struct_asym.id
_struct_asym.entity_id
_struct_asym.details
A 1 foo
"""


class Tests(unittest.TestCase):

    def test_old_file_read_default(self):
        """Test default handling of old files"""
        cif = """
loop_
_audit_conform.dict_name
_audit_conform.dict_version
mmcif_pdbx.dic  5.311
mmcif_ma.dic    0.14
"""
        s, = modelcif.reader.read(StringIO(cif))

    def test_old_file_read_fail(self):
        """Test failure reading old files"""
        cif = """
loop_
_audit_conform.dict_name
_audit_conform.dict_version
mmcif_pdbx.dic  5.311
mmcif_ma.dic    0.1.3
"""
        self.assertRaises(modelcif.reader.OldFileError,
                          modelcif.reader.read, StringIO(cif),
                          reject_old_file=True)

    def test_new_file_read_ok(self):
        """Test success reading not-old files"""
        # File read is OK if version is new enough, or version cannot be parsed
        # because it is non-int or has too many elements
        for ver in ('1.3.2', '1.3', '0.0.4.3', '0.0a'):
            cif = """
loop_
_audit_conform.dict_name
_audit_conform.dict_version
mmcif_pdbx.dic  5.311
mmcif_ma.dic    %s
""" % ver
            s, = modelcif.reader.read(StringIO(cif), reject_old_file=True)

    def test_software_group_handler(self):
        """Test SoftwareGroupHandler and SoftwareParameterHandler"""
        cif = """
loop_
_ma_software_parameter.parameter_id
_ma_software_parameter.group_id
_ma_software_parameter.data_type
_ma_software_parameter.name
_ma_software_parameter.value
_ma_software_parameter.description
1 1 integer foo 42 foodesc
2 1 boolean bar YES .
3 1 string baz ok .
4 1 integer-csv intlist 1,2,3,4 .
5 1 float-csv floatlist 1.5,3.8 .
#
loop_
_ma_software_group.ordinal_id
_ma_software_group.group_id
_ma_software_group.software_id
_ma_software_group.parameter_group_id
1 1 1 .
2 1 2 .
3 2 3 .
4 2 4 1
"""
        s, = modelcif.reader.read(StringIO(cif))
        s1, s2, s3, s4 = s.software
        g1, g2 = s.software_groups
        self.assertEqual(len(g1), 2)
        self.assertEqual(len(g2), 2)
        self.assertIsInstance(g1[0], modelcif.Software)
        self.assertIsInstance(g1[1], modelcif.Software)
        self.assertEqual(g1[0], s1)
        self.assertEqual(g1[1], s2)

        self.assertIsInstance(g2[0], modelcif.Software)
        self.assertIsInstance(g2[1], modelcif.SoftwareWithParameters)
        self.assertEqual(g2[0], s3)
        self.assertEqual(g2[1].software, s4)
        p1, p2, p3, intlist, floatlist = g2[1].parameters
        self.assertEqual(p1.name, 'foo')
        self.assertEqual(p1.value, 42)
        self.assertEqual(p1.description, 'foodesc')
        self.assertEqual(p2.name, 'bar')
        self.assertTrue(p2.value)
        self.assertIsNone(p2.description)
        self.assertEqual(p3.name, 'baz')
        self.assertEqual(p3.value, 'ok')
        self.assertIsNone(p3.description)
        self.assertEqual(intlist.value, [1, 2, 3, 4])
        f1, f2 = floatlist.value
        self.assertAlmostEqual(f1, 1.5, delta=1e-1)
        self.assertAlmostEqual(f2, 3.8, delta=1e-1)

    def test_atom_id_map(self):
        """Test _AtomIDMap class"""
        m = modelcif.reader._AtomIDMap()
        # Should be a noop
        m.add(atom_id='non-numeric', model_num='1')
        m.add(atom_id='42', model_num='1')
        m.add(atom_id='48', model_num='1')
        m.add(atom_id='24', model_num='1')
        self.assertEqual(m._model_to_id_range['1'], [24, 48])
        self.assertEqual(m.get(24), '1')
        self.assertEqual(m.get(34), '1')
        self.assertRaises(ValueError, m.get, 99)

    def test_enumeration_mapper(self):
        """Test EnumerationMapper class"""
        m = modelcif.reader._EnumerationMapper(
            modelcif.reference, modelcif.reference.TargetReference)
        # Check get of a handled enumeration value
        unp = m.get('UNP', None)
        self.assertIs(unp, modelcif.reference.UniProt)
        self.assertEqual(unp.name, 'UNP')
        self.assertIsNone(unp.other_details)
        # We should get the same class each time (case insensitive)
        unp2 = m.get('unp', None)
        self.assertIs(unp, unp2)
        # Check get of an unhandled value
        miss = m.get('MIS', None)
        self.assertEqual(miss.name, 'MIS')
        self.assertIsNone(unp.other_details)
        # We should get the same class each time (case insensitive)
        miss2 = m.get('mis', None)
        self.assertIs(miss, miss2)
        # Check get of a custom "other" value
        custom = m.get('other', "custom type 1")
        self.assertEqual(custom.name, 'Other')
        self.assertEqual(custom.other_details, "custom type 1")
        # We should get the same class each time (case insensitive)
        custom2 = m.get('Other', "CUSTOM TYPE 1")
        self.assertIs(custom, custom2)
        # Check get of a different custom "other" value
        custom = m.get('other', "custom type 2")
        self.assertEqual(custom.name, 'Other')
        self.assertEqual(custom.other_details, "custom type 2")

    def test_database_handler(self):
        """Test DatabaseHandler"""
        cif = """
_database_2.database_id                'PDB'
_database_2.database_code              '5HVP'
"""
        s, = modelcif.reader.read(StringIO(cif))
        self.assertEqual(s.database.id, 'PDB')
        self.assertEqual(s.database.code, '5HVP')

    def test_target_ref_db_handler(self):
        """Test TargetRefDBHander"""
        cif = """
loop_
_ma_target_ref_db_details.target_entity_id
_ma_target_ref_db_details.db_name
_ma_target_ref_db_details.db_name_other_details
_ma_target_ref_db_details.db_code
_ma_target_ref_db_details.db_accession
_ma_target_ref_db_details.seq_db_isoform
_ma_target_ref_db_details.seq_db_align_begin
_ma_target_ref_db_details.seq_db_align_end
_ma_target_ref_db_details.ncbi_taxonomy_id
_ma_target_ref_db_details.organism_scientific
_ma_target_ref_db_details.seq_db_sequence_version_date
_ma_target_ref_db_details.seq_db_sequence_checksum
_ma_target_ref_db_details.is_primary
1 UNP . MED1_YEAST Q12321 test_iso 1 10 test_tax test_org 1996-11-01
637FEA3E78D915BC YES
1 Other foo . . ? 1 10 . . . . NO
1 other bar . . ? 1 10 . . . . .
1 MIS baz . . ? 1 10 . . . . .
"""
        s, = modelcif.reader.read(StringIO(cif))
        e, = s.entities
        r1, r2, r3, r4 = e.references
        self.assertIsInstance(r1, modelcif.reference.UniProt)
        self.assertEqual(r1.code, 'MED1_YEAST')
        self.assertEqual(r1.accession, 'Q12321')
        self.assertEqual(r1.isoform, 'test_iso')
        self.assertEqual(r1.align_begin, 1)
        self.assertEqual(r1.align_end, 10)
        self.assertEqual(r1.ncbi_taxonomy_id, 'test_tax')
        self.assertEqual(r1.organism_scientific, 'test_org')
        self.assertEqual(r1.sequence_version_date, date(1996, 11, 1))
        self.assertIsNone(r1.sequence)
        self.assertIsNone(r1.details)
        self.assertTrue(r1.is_primary)
        self.assertEqual(r1.alignments, [])
        self.assertEqual(r2.name, 'Other')
        self.assertFalse(r2.is_primary)
        self.assertEqual(r2.other_details, 'foo')
        self.assertEqual(r3.name, 'Other')
        self.assertEqual(r3.other_details, 'bar')
        self.assertIsNone(r3.is_primary)
        self.assertEqual(r4.name, 'MIS')
        self.assertIsNone(r4.other_details)  # should be ignored

    def test_target_ref_db_handler_with_struct_ref(self):
        """Test TargetRefDBHander combined with struct_ref info"""
        cif = """
loop_
_struct_ref.id
_struct_ref.entity_id
_struct_ref.db_name
_struct_ref.db_code
_struct_ref.pdbx_db_accession
_struct_ref.pdbx_align_begin
_struct_ref.pdbx_seq_one_letter_code
_struct_ref.details
1 1 UNP MED1_YEAST Q12321 1 DSYVETLDCC "test details"
2 1 UNP sr_only_code sr_only_acc 1 DSYVETLDPP .
#
#
loop_
_struct_ref_seq.align_id
_struct_ref_seq.ref_id
_struct_ref_seq.seq_align_beg
_struct_ref_seq.seq_align_end
_struct_ref_seq.db_align_beg
_struct_ref_seq.db_align_end
1 1 1 10 1 10
2 2 1 10 1 10
#
loop_
_ma_target_ref_db_details.target_entity_id
_ma_target_ref_db_details.db_name
_ma_target_ref_db_details.db_name_other_details
_ma_target_ref_db_details.db_code
_ma_target_ref_db_details.db_accession
_ma_target_ref_db_details.seq_db_isoform
_ma_target_ref_db_details.seq_db_align_begin
_ma_target_ref_db_details.seq_db_align_end
_ma_target_ref_db_details.ncbi_taxonomy_id
_ma_target_ref_db_details.organism_scientific
_ma_target_ref_db_details.seq_db_sequence_version_date
_ma_target_ref_db_details.seq_db_sequence_checksum
_ma_target_ref_db_details.is_primary
1 UNP . MED1_YEAST Q12321 test_iso 1 10 test_tax test_org 1996-11-01
637FEA3E78D915BC YES
1 UNP . rd_only_code rd_only_acc rd_only_iso . . . . . . NO
"""
        s, = modelcif.reader.read(StringIO(cif))
        e, = s.entities
        r1, r2, r3 = e.references
        # r1 should contain both target_ref_db and struct_ref info
        self.assertIsInstance(r1, modelcif.reference.UniProt)
        self.assertEqual(r1.code, 'MED1_YEAST')
        self.assertEqual(r1.accession, 'Q12321')
        self.assertEqual(r1.isoform, 'test_iso')
        self.assertEqual(r1.align_begin, 1)
        self.assertEqual(r1.align_end, 10)
        self.assertEqual(r1.ncbi_taxonomy_id, 'test_tax')
        self.assertEqual(r1.organism_scientific, 'test_org')
        self.assertEqual(r1.sequence_version_date, date(1996, 11, 1))
        self.assertEqual(r1.sequence, 'DSYVETLDCC')
        self.assertEqual(r1.details, "test details")
        self.assertTrue(r1.is_primary)
        a, = r1.alignments
        self.assertEqual(a.db_begin, 1)
        self.assertEqual(a.db_end, 10)
        self.assertEqual(a.entity_begin, 1)
        self.assertEqual(a.entity_end, 10)
        # r2 should contain only target_ref_db info
        self.assertIsInstance(r2, modelcif.reference.UniProt)
        self.assertEqual(r2.code, 'rd_only_code')
        self.assertEqual(r2.accession, 'rd_only_acc')
        self.assertEqual(r2.isoform, 'rd_only_iso')
        self.assertIsNone(r2.sequence)
        self.assertFalse(r2.is_primary)
        # r3 should contain only struct_ref info
        self.assertIsInstance(r3, modelcif.reference.UniProt)
        self.assertEqual(r3.code, 'sr_only_code')
        self.assertEqual(r3.accession, 'sr_only_acc')
        self.assertIsNone(r3.isoform)
        self.assertIsNone(r3.ncbi_taxonomy_id)
        self.assertEqual(r3.sequence, 'DSYVETLDPP')
        self.assertIsNone(r3.is_primary)
        a, = r3.alignments
        self.assertEqual(a.db_begin, 1)
        self.assertEqual(a.db_end, 10)
        self.assertEqual(a.entity_begin, 1)
        self.assertEqual(a.entity_end, 10)

    def test_transformation_handler(self):
        """Test _TransformationHandler"""
        cif = """
loop_
_ma_template_trans_matrix.id
_ma_template_trans_matrix.rot_matrix[1][1]
_ma_template_trans_matrix.rot_matrix[2][1]
_ma_template_trans_matrix.rot_matrix[3][1]
_ma_template_trans_matrix.rot_matrix[1][2]
_ma_template_trans_matrix.rot_matrix[2][2]
_ma_template_trans_matrix.rot_matrix[3][2]
_ma_template_trans_matrix.rot_matrix[1][3]
_ma_template_trans_matrix.rot_matrix[2][3]
_ma_template_trans_matrix.rot_matrix[3][3]
_ma_template_trans_matrix.tr_vector[1]
_ma_template_trans_matrix.tr_vector[2]
_ma_template_trans_matrix.tr_vector[3]
1 1.000000 0.000000 0.000000 0.000000 1.000000 0.000000 0.000000 0.000000
1.000000 0.000 0.000 0.000
"""
        s, = modelcif.reader.read(StringIO(cif))
        t, = s.template_transformations
        self.assertAlmostEqual(t.rot_matrix[0][0], 1.0, delta=1e-6)
        self.assertAlmostEqual(t.tr_vector[0], 0.0, delta=1e-6)

    def test_template_details_handler(self):
        """Test _TemplateDetailsHandler"""
        cif = """
loop_
_ma_template_details.ordinal_id
_ma_template_details.template_id
_ma_template_details.template_origin
_ma_template_details.template_entity_type
_ma_template_details.template_trans_matrix_id
_ma_template_details.template_data_id
_ma_template_details.target_asym_id
_ma_template_details.template_label_asym_id
_ma_template_details.template_label_entity_id
_ma_template_details.template_model_num
_ma_template_details.template_auth_asym_id
1 1 'reference database' polymer 1 2 A B 3 4 Z
2 2 'reference database' polymer 2 3 . B 3 4 Z
3 3 customized polymer 2 3 . B 3 4 Z
"""
        s, = modelcif.reader.read(StringIO(cif))
        t, t2, t3 = s.templates
        self.assertIsInstance(t, modelcif.Template)
        self.assertIsInstance(t2, modelcif.Template)
        self.assertIsInstance(t3, modelcif.CustomTemplate)
        self.assertEqual(t.entity_id, '3')
        self.assertEqual(t.model_num, 4)
        self.assertEqual(t.asym_id, 'B')
        self.assertEqual(t.strand_id, 'Z')
        self.assertEqual(len(s.alignments), 0)

    def test_template_customized_handler(self):
        """Test _TemplateCustomizedHandler"""
        cif = """
loop_
_ma_template_details.ordinal_id
_ma_template_details.template_id
1 1
#
loop_
_ma_template_customized.template_id
_ma_template_customized.details
1 'details x'
2 'details y'
"""
        s, = modelcif.reader.read(StringIO(cif))
        t1, t2 = s.templates
        # template_details does not specify template_origin, so template #1
        # will be initially instantiated as a Template, and should be corrected
        # to CustomTemplate on reading template_customized:
        self.assertIsInstance(t1, modelcif.CustomTemplate)
        self.assertEqual(t1.details, 'details x')
        self.assertEqual(len(t1.atoms), 0)
        self.assertIsInstance(t2, modelcif.CustomTemplate)
        self.assertEqual(t2.details, 'details y')
        self.assertEqual(len(t2.atoms), 0)

    def test_template_details_handler_nonpoly(self):
        """Test _TemplateDetailsHandler with nonpolymeric template"""
        cif = """
loop_
_pdbx_entity_nonpoly.entity_id
_pdbx_entity_nonpoly.name
_pdbx_entity_nonpoly.comp_id
_pdbx_entity_nonpoly.ma_model_mode
3 Heme HEM explicit
#
loop_
_ma_template_details.ordinal_id
_ma_template_details.template_id
_ma_template_details.template_origin
_ma_template_details.template_entity_type
_ma_template_details.template_trans_matrix_id
_ma_template_details.template_data_id
_ma_template_details.target_asym_id
_ma_template_details.template_label_asym_id
_ma_template_details.template_label_entity_id
_ma_template_details.template_model_num
_ma_template_details.template_auth_asym_id
1 1 'reference database' non-polymer 1 2 A B 3 4 Z
#
loop_
_ma_template_non_poly.template_id
_ma_template_non_poly.comp_id
_ma_template_non_poly.details
1 HEM "Template Heme"
"""
        s, = modelcif.reader.read(StringIO(cif))
        t, = s.templates
        self.assertEqual(t.entity_id, '3')
        self.assertEqual(t.model_num, 4)
        self.assertEqual(t.asym_id, 'B')
        self.assertEqual(t.strand_id, 'Z')
        self.assertEqual(len(s.alignments), 0)
        self.assertEqual(t.entity.description, 'Template Heme')
        a, = s.asym_units
        self.assertIsInstance(a, modelcif.NonPolymerFromTemplate)
        self.assertIs(a.template, t)
        self.assertTrue(a.explicit)

    def test_custom_template_coord_handler(self):
        """Test reading of coordinates for CustomTemplate"""
        cif = """
loop_
_ma_template_details.ordinal_id
_ma_template_details.template_id
_ma_template_details.template_origin
_ma_template_details.template_entity_type
_ma_template_details.template_trans_matrix_id
_ma_template_details.template_data_id
_ma_template_details.target_asym_id
_ma_template_details.template_label_asym_id
_ma_template_details.template_label_entity_id
_ma_template_details.template_model_num
_ma_template_details.template_auth_asym_id
1 1 customized polymer 2 3 . B 3 4 Z
#
loop_
_ma_template_customized.template_id
_ma_template_customized.details
1 'Provided by user'
#
loop_
_ma_template_coord.template_id
_ma_template_coord.group_PDB
_ma_template_coord.ordinal_id
_ma_template_coord.type_symbol
_ma_template_coord.label_atom_id
_ma_template_coord.label_comp_id
_ma_template_coord.label_seq_id
_ma_template_coord.label_asym_id
_ma_template_coord.auth_seq_id
_ma_template_coord.auth_asym_id
_ma_template_coord.auth_atom_id
_ma_template_coord.auth_comp_id
_ma_template_coord.Cartn_x
_ma_template_coord.Cartn_y
_ma_template_coord.Cartn_z
_ma_template_coord.occupancy
_ma_template_coord.label_entity_id
_ma_template_coord.B_iso_or_equiv
_ma_template_coord.formal_charge
1 ATOM 1 C CA ALA 1 A 42 A X XXX 0 1.000 2.000 0.500 9 2.000 1.000
1 ATOM 2 O OXT CYS 2 A . A . . 1.000 2.000 3.000 . 9 . .
#
"""
        s, = modelcif.reader.read(StringIO(cif))
        t, = s.templates
        self.assertIsInstance(t, modelcif.CustomTemplate)
        self.assertEqual(t.details, 'Provided by user')
        self.assertEqual(len(t.atoms), 2)
        a1 = t.atoms[0]
        self.assertEqual(a1.seq_id, 1)
        self.assertEqual(a1.atom_id, 'CA')
        self.assertEqual(a1.type_symbol, 'C')
        self.assertAlmostEqual(a1.x, 0.0, delta=1e-2)
        self.assertAlmostEqual(a1.y, 1.0, delta=1e-2)
        self.assertAlmostEqual(a1.z, 2.0, delta=1e-2)
        self.assertAlmostEqual(a1.occupancy, 0.5, delta=1e-2)
        self.assertAlmostEqual(a1.biso, 2.0, delta=1e-2)
        self.assertAlmostEqual(a1.charge, 1.0, delta=1e-2)
        self.assertEqual(a1.auth_seq_id, 42)
        self.assertEqual(a1.auth_comp_id, 'XXX')
        self.assertEqual(a1.auth_atom_id, 'X')

        a2 = t.atoms[1]
        self.assertEqual(a2.seq_id, 2)
        self.assertEqual(a2.atom_id, 'OXT')
        self.assertEqual(a2.type_symbol, 'O')

    def test_entity_nonpoly_bad_model_mode(self):
        """Test pdbx_entity_nonpoly with missing ma_model_mode"""
        cif = """
loop_
_struct_asym.id
_struct_asym.entity_id
A 1
B 2
C 3
#
loop_
_pdbx_entity_nonpoly.entity_id
_pdbx_entity_nonpoly.name
_pdbx_entity_nonpoly.comp_id
_pdbx_entity_nonpoly.ma_model_mode
1 test1 TE1 explicit
2 test2 TE2 .
3 test3 TE3 ?
#
loop_
_ma_template_details.ordinal_id
_ma_template_details.template_id
_ma_template_details.template_origin
_ma_template_details.template_entity_type
_ma_template_details.template_trans_matrix_id
_ma_template_details.template_data_id
_ma_template_details.target_asym_id
_ma_template_details.template_label_asym_id
_ma_template_details.template_label_entity_id
_ma_template_details.template_model_num
_ma_template_details.template_auth_asym_id
1 1 'reference database' non-polymer 1 2 A A 1 4 Z
2 2 'reference database' non-polymer 1 3 B B 2 4 Z
3 3 'reference database' non-polymer 1 4 C C 3 4 Z
#
loop_
_ma_template_non_poly.template_id
_ma_template_non_poly.comp_id
_ma_template_non_poly.details
1 TE1 test1
2 TE2 test2
3 TE3 test3
"""
        s, = modelcif.reader.read(StringIO(cif))
        a1, a2, a3 = s.asym_units
        self.assertTrue(a1.explicit)
        self.assertIsNone(a2.explicit)
        self.assertIs(a3.explicit, ihm.unknown)

    def test_template_ref_db_handler(self):
        """Test _TemplateRefDBHandler"""
        cif = """
loop_
_ma_template_ref_db_details.template_id
_ma_template_ref_db_details.db_name
_ma_template_ref_db_details.db_name_other_details
_ma_template_ref_db_details.db_accession_code
_ma_template_ref_db_details.db_version_date
1 PDB . 3nc1 2021-10-06
1 MIS . testacc .
1 Other foo acc2 .
1 PubChem . 1046 .
1 AlphaFoldDB . I6XD65 2022-06-01
"""
        s, = modelcif.reader.read(StringIO(cif))
        t, = s.templates
        r1, r2, r3, r4, r5 = t.references
        self.assertIsInstance(r1, modelcif.reference.PDB)
        self.assertEqual(r1.accession, '3nc1')
        self.assertEqual(r1.db_version_date, date(2021, 10, 6))
        self.assertEqual(r2.name, 'MIS')
        self.assertIsNone(r2.other_details)
        self.assertIsNone(r2.db_version_date)
        self.assertEqual(r3.name, 'Other')
        self.assertEqual(r3.other_details, 'foo')
        self.assertIsInstance(r4, modelcif.reference.PubChem)
        self.assertEqual(r4.accession, '1046')
        self.assertIsInstance(r5, modelcif.reference.AlphaFoldDB)
        self.assertEqual(r5.accession, 'I6XD65')
        self.assertEqual(r5.db_version_date, date(2022, 6, 1))

    def _get_models_cif(self, old=False):
        if old:
            cif = """
loop_
_ma_model_list.ordinal_id
_ma_model_list.model_id
_ma_model_list.model_group_id
_ma_model_list.model_name
_ma_model_list.model_group_name
_ma_model_list.assembly_id
_ma_model_list.data_id
_ma_model_list.model_type
_ma_model_list.model_type_other_details
1 1 1 'Best scoring model' 'All models' 99 4 'Homology model' .
2 2 1 '2nd best scoring model' 'All models' 99 5 'Ab initio model' .
3 3 2 'Best scoring model' 'group2' 99 6 'Other' 'Custom other model'
#
"""
        else:
            cif = """
loop_
_ma_model_list.ordinal_id
_ma_model_list.model_name
_ma_model_list.assembly_id
_ma_model_list.data_id
_ma_model_list.model_type
_ma_model_list.model_type_other_details
1 'Best scoring model' 99 4 'Homology model' .
2 '2nd best scoring model' 99 5 'Ab initio model' .
3 'Best scoring model' 99 6 'Other' 'Custom other model'
#
loop_
_ma_model_group.id
_ma_model_group.name
_ma_model_group.details
1 'All models' .
2 'group2' 'second group details'
#
#
loop_
_ma_model_group_link.group_id
_ma_model_group_link.model_id
1 1
1 2
2 3
#
"""
        cif += """
loop_
_atom_site.group_PDB
_atom_site.id
_atom_site.type_symbol
_atom_site.label_atom_id
_atom_site.label_alt_id
_atom_site.label_comp_id
_atom_site.label_seq_id
_atom_site.auth_seq_id
_atom_site.pdbx_PDB_ins_code
_atom_site.label_asym_id
_atom_site.Cartn_x
_atom_site.Cartn_y
_atom_site.Cartn_z
_atom_site.occupancy
_atom_site.label_entity_id
_atom_site.auth_asym_id
_atom_site.B_iso_or_equiv
_atom_site.pdbx_PDB_model_num
ATOM 1 C CA . ASP 1 1 ? A 1.000 2.000 3.000 . 1 A . 6
ATOM 2 C CA . ASP 1 1 ? A 1.000 2.000 3.000 . 1 A . 8
"""
        return cif

    def test_model_list_handler_default_old(self):
        """Test _ModelListHandler with default model class, old dictionary"""
        self._test_model_list_handler_default(old=True)

    def test_model_list_handler_default(self):
        """Test _ModelListHandler with default model class"""
        self._test_model_list_handler_default(old=False)

    def _test_model_list_handler_default(self, old):
        cif = self._get_models_cif(old=old)
        s, = modelcif.reader.read(StringIO(cif))
        mg1, mg2, mg3 = s.model_groups
        self.assertEqual(mg1.name, 'All models')
        m1, m2 = list(mg1)
        self.assertIsInstance(m1, modelcif.model.HomologyModel)
        self.assertEqual(m1.model_type, 'Homology model')
        self.assertIsNone(m1.other_details)
        self.assertIsInstance(m2, modelcif.model.AbInitioModel)
        self.assertEqual(m2.model_type, 'Ab initio model')
        self.assertIsNone(m2.other_details)
        self.assertEqual(m1.name, 'Best scoring model')
        self.assertEqual(m2.name, '2nd best scoring model')
        self.assertEqual(mg2.name, 'group2')
        m1, = list(mg2)
        self.assertEqual(m1.model_type, 'Other')
        self.assertEqual(m1.other_details, 'Custom other model')
        self.assertEqual(m1.name, 'Best scoring model')
        self.assertEqual(m1.assembly._id, '99')
        # Last group is auto-created to contain the non-grouped models
        # referenced by atom_site
        self.assertIsNone(mg3.name)
        m1, m2 = list(mg3)
        self.assertEqual(m1.model_type, 'Other')
        self.assertEqual(m1._id, '6')
        self.assertEqual(m2.model_type, 'Other')
        self.assertEqual(m2._id, '8')

    def test_model_list_handler_group_new_old(self):
        """Test _ModelListHandler handling mix of new and old style groups"""
        cif = """
loop_
_ma_model_list.ordinal_id
_ma_model_list.model_id
_ma_model_list.model_group_id
_ma_model_list.model_name
_ma_model_list.model_group_name
_ma_model_list.data_id
_ma_model_list.model_type
_ma_model_list.model_type_other_details
1 1 1 . . 4 'Homology model' .
2 2 . . . 4 'Homology model' .
3 3 1 . . 4 'Homology model' .
#
#
loop_
_ma_model_group.id
_ma_model_group.name
_ma_model_group.details
1 'group1' .
2 'group2' .
#
#
loop_
_ma_model_group_link.model_id
_ma_model_group_link.group_id
2 1
3 2
"""
        s, = modelcif.reader.read(StringIO(cif))
        # model1 is in group1, using old-style tables;
        # model2 is in group1, using new-style tables;
        # model3 is in group2 according to new-style tables but group1
        # according to old style (new-style should take precedence)
        mg1, mg2 = s.model_groups
        self.assertEqual(mg1._id, '1')
        self.assertEqual(mg2._id, '2')
        self.assertEqual([m._id for m in mg1], ['2', '1'])
        self.assertEqual([m._id for m in mg2], ['3'])

    def test_model_list_handler_custom(self):
        """Test _ModelListHandler with custom model class"""
        class MyModel(modelcif.model.Model):
            """Custom model type"""
            pass
        cif = self._get_models_cif()
        s, = modelcif.reader.read(StringIO(cif), model_class=MyModel)
        mg1, mg2, mg3 = s.model_groups
        m1, m2 = list(mg1)
        m3, = list(mg2)
        m4, m5 = list(mg3)
        # Custom model type should always be returned, regardless of what
        # the mmCIF file says it is, but model_type should be set
        self.assertIsInstance(m1, MyModel)
        self.assertIsInstance(m2, MyModel)
        self.assertIsInstance(m3, MyModel)
        self.assertIsInstance(m4, MyModel)
        self.assertIsInstance(m5, MyModel)
        self.assertEqual(m1.model_type, 'Homology model')
        self.assertEqual(m2.model_type, 'Ab initio model')
        self.assertEqual(m3.model_type, 'Other')
        self.assertEqual(m4.model_type, 'Other')
        self.assertEqual(m5.model_type, 'Other')

    def test_assembly_handler(self):
        """Test _AssemblyHandler and _AssemblyDetailsHandler"""
        cif = """
loop_
_entity_poly_seq.entity_id
_entity_poly_seq.num
_entity_poly_seq.mon_id
1 1 ALA
1 2 ALA
#
loop_
_struct_asym.id
_struct_asym.entity_id
_struct_asym.details
A 1 Nup84
#
loop_
_ma_struct_assembly.ordinal_id
_ma_struct_assembly.assembly_id
_ma_struct_assembly.entity_id
_ma_struct_assembly.asym_id
_ma_struct_assembly.seq_id_begin
_ma_struct_assembly.seq_id_end
1 1 1 A 1 2
2 1 1 A 1 1
3 1 1 A . .
#
loop_
_ma_struct_assembly_details.assembly_id
_ma_struct_assembly_details.assembly_name
_ma_struct_assembly_details.assembly_description
1 foo bar
"""
        s, = modelcif.reader.read(StringIO(cif))
        a, = s.assemblies
        self.assertEqual(a.name, 'foo')
        self.assertEqual(a.description, 'bar')
        self.assertEqual(len(a), 3)
        # Complete asym
        self.assertIsInstance(a[0], modelcif.AsymUnit)
        # asym range
        self.assertIsInstance(a[1], modelcif.AsymUnitRange)
        self.assertEqual(a[1].seq_id_range, (1, 1))
        # No specified range -> complete asym
        self.assertIsInstance(a[2], modelcif.AsymUnit)

    def test_template_poly_segment_handler(self):
        """Test _TemplatePolySegmentHandler"""
        cif = """
loop_
_ma_template_poly_segment.id
_ma_template_poly_segment.template_id
_ma_template_poly_segment.residue_number_begin
_ma_template_poly_segment.residue_number_end
1 42 2 9
"""
        s, = modelcif.reader.read(StringIO(cif))
        seg, = s.template_segments
        self.assertEqual(seg.template._id, '42')
        self.assertEqual(seg.seq_id_range, (2, 9))

    def test_data__handler(self):
        """Test _DataHandler"""
        cif = """
loop_
_ma_data.id
_ma_data.name
_ma_data.content_type
_ma_data.content_type_other_details
1 'Template Structure' 'template structure' .
2 'Model subunit' target .
3 'Default model name' 'model coordinates' .
#
loop_
_ma_template_details.ordinal_id
_ma_template_details.template_id
_ma_template_details.template_origin
_ma_template_details.template_entity_type
_ma_template_details.template_trans_matrix_id
_ma_template_details.template_data_id
_ma_template_details.target_asym_id
_ma_template_details.template_label_asym_id
_ma_template_details.template_label_entity_id
_ma_template_details.template_model_num
_ma_template_details.template_auth_asym_id
1 1 'reference database' polymer 1 1 A A 1 1 Z
#
loop_
_ma_model_list.ordinal_id
_ma_model_list.model_id
_ma_model_list.model_group_id
_ma_model_list.model_name
_ma_model_list.model_group_name
_ma_model_list.assembly_id
_ma_model_list.data_id
_ma_model_list.model_type
_ma_model_list.model_type_other_details
1 1 1 'Model name' 'All models' 1 3 'Homology model' .
"""
        s, = modelcif.reader.read(StringIO(cif))
        d1, d2, d3 = s.data
        self.assertIsInstance(d1, modelcif.Template)
        # d2 is not referenced by any other table, so gets Data base class
        self.assertIsInstance(d2, modelcif.data.Data)
        self.assertIsInstance(d3, modelcif.model.Model)
        # Name not given in template_details so taken from ma_data
        self.assertEqual(d1.name, 'Template Structure')
        self.assertEqual(d2.name, 'Model subunit')
        # Name in model_list used rather than that from ma_data
        self.assertEqual(d3.name, 'Model name')

    def test_data_group_handler(self):
        """Test _DataGroupHandler"""
        cif = """
loop_
_ma_data.id
_ma_data.name
_ma_data.content_type
_ma_data.content_type_other_details
1 'Template Structure' 'template structure' .
2 'Model subunit' target .
#
loop_
_ma_template_details.ordinal_id
_ma_template_details.template_id
_ma_template_details.template_origin
_ma_template_details.template_entity_type
_ma_template_details.template_trans_matrix_id
_ma_template_details.template_data_id
_ma_template_details.target_asym_id
_ma_template_details.template_label_asym_id
_ma_template_details.template_label_entity_id
_ma_template_details.template_model_num
_ma_template_details.template_auth_asym_id
1 1 'reference database' polymer 1 1 A A 1 1 Z
#
loop_
_ma_data_group.ordinal_id
_ma_data_group.group_id
_ma_data_group.data_id
1 1 1
2 1 2
3 2 3
"""
        s, = modelcif.reader.read(StringIO(cif))
        g1, g2, = s.data_groups
        self.assertEqual(len(g1), 2)
        self.assertIsInstance(g1[0], modelcif.Template)
        self.assertEqual(g1[0]._data_id, '1')
        self.assertIsInstance(g1[1], modelcif.data.Data)
        self.assertEqual(g1[1]._data_id, '2')
        self.assertEqual(len(g2), 1)
        self.assertIsNone(g2[0])

    def test_data_ref_db_handler(self):
        """Test _DataRefDBHandler"""
        cif = """
loop_
_ma_data.id
_ma_data.name
_ma_data.content_type
_ma_data.content_type_other_details
1 defaultname1 'reference database' .
2 defaultname2 'reference database' .
#
loop_
_ma_data_ref_db.data_id
_ma_data_ref_db.name
_ma_data_ref_db.location_url
_ma_data_ref_db.version
_ma_data_ref_db.release_date
1 name1 url1 1.0 1979-11-22
2 . url2 . .
"""
        s, = modelcif.reader.read(StringIO(cif))
        d1, d2 = s.data
        self.assertIsInstance(d1, modelcif.ReferenceDatabase)
        self.assertIsInstance(d2, modelcif.ReferenceDatabase)
        # Name in ma_data_ref_db used rather than that from ma_data
        self.assertEqual(d1.name, 'name1')
        self.assertEqual(d1.url, 'url1')
        self.assertEqual(d1.version, '1.0')
        self.assertIsInstance(d1.release_date, date)
        self.assertEqual(d1.release_date, date(1979, 11, 22))
        # Name not given in ma_data_ref_db so taken from ma_data
        self.assertEqual(d2.name, 'defaultname2')
        self.assertIsNone(d2.version)
        self.assertIsNone(d2.release_date)

    def test_protocol_handler(self):
        """Test _ProtocolHandler"""
        cif = """
loop_
_ma_protocol_step.ordinal_id
_ma_protocol_step.protocol_id
_ma_protocol_step.step_id
_ma_protocol_step.method_type
_ma_protocol_step.step_name
_ma_protocol_step.details
_ma_protocol_step.software_group_id
_ma_protocol_step.input_data_group_id
_ma_protocol_step.output_data_group_id
1 1 1 'template search' 'ModPipe Seq-Prf (0001)' . 1 1 2
2 1 2 'template selection' . . . . .
3 1 3 'target-template alignment' . . . . .
4 1 4 modeling . . 2 2 1
5 1 5 'model selection' . . 1 1 1
6 1 6 'model refinement' . . . . .
7 1 7 other testname testdetails 42 99 66
8 1 8 'model quality assessment' . . . . .
9 1 9 'backbone generation' . . . . .
10 1 10 'sequence design' . . . . .
11 1 11 'structure validation' . . . . .
12 1 12 'energy estimate' . . . . .
13 1 13 'experimental validation' . . . . .
"""
        s, = modelcif.reader.read(StringIO(cif))
        p, = s.protocols
        self.assertEqual(len(p.steps), 13)
        self.assertIsInstance(p.steps[0], modelcif.protocol.TemplateSearchStep)
        self.assertIsInstance(p.steps[1],
                              modelcif.protocol.TemplateSelectionStep)
        self.assertIsInstance(p.steps[2],
                              modelcif.protocol.TargetTemplateAlignmentStep)
        self.assertIsInstance(p.steps[3], modelcif.protocol.ModelingStep)
        self.assertIsInstance(p.steps[4], modelcif.protocol.ModelSelectionStep)
        self.assertIsInstance(p.steps[5],
                              modelcif.protocol.ModelRefinementStep)
        self.assertIsInstance(p.steps[6], modelcif.protocol.Step)
        s7 = p.steps[6]
        self.assertEqual(s7.method_type, "other")
        self.assertEqual(s7.name, "testname")
        self.assertEqual(s7.details, "testdetails")
        self.assertEqual(s7.input_data._id, '99')
        self.assertEqual(s7.output_data._id, '66')
        self.assertEqual(s7.software._id, '42')
        self.assertIsInstance(p.steps[7],
                              modelcif.protocol.ModelAssessmentStep)
        self.assertIsInstance(p.steps[8],
                              modelcif.protocol.BackboneGenerationStep)
        self.assertIsInstance(p.steps[9],
                              modelcif.protocol.SequenceDesignStep)
        self.assertIsInstance(p.steps[10],
                              modelcif.protocol.StructureValidationStep)
        self.assertIsInstance(p.steps[11],
                              modelcif.protocol.EnergyEstimateStep)
        self.assertIsInstance(p.steps[12],
                              modelcif.protocol.ExperimentalValidationStep)

    def test_target_entity_handler(self):
        """Test _TargetEntityHandler"""
        cif = """
loop_
_ma_target_entity.entity_id
_ma_target_entity.data_id
_ma_target_entity.origin
1 2 'reference database'
"""
        s, = modelcif.reader.read(StringIO(cif))
        e, = s.entities
        self.assertEqual(e._data_id, '2')

    def test_qa_metric_global_handler(self):
        """Test _QAMetricGlobalHandler"""
        cif = """
loop_
_ma_model_list.ordinal_id
_ma_model_list.model_id
_ma_model_list.model_group_id
_ma_model_list.model_name
_ma_model_list.model_group_name
_ma_model_list.assembly_id
_ma_model_list.data_id
_ma_model_list.model_type
_ma_model_list.model_type_other_details
1 1 1 'Best scoring model' 'All models' 1 4 'Homology model' .
#
loop_
_ma_qa_metric.id
_ma_qa_metric.name
_ma_qa_metric.description
_ma_qa_metric.type
_ma_qa_metric.mode
_ma_qa_metric.type_other_details
_ma_qa_metric.software_group_id
1 MPQS 'ModPipe Quality Score' other global
'composite score, values >1.1 are considered reliable' 1
2 zDOPE 'Normalized DOPE' zscore global . 2
3 'TSVMod RMSD' 'TSVMod predicted RMSD (MSALL)' distance global . .
4 'TSVMod NO35' . 'normalized score' global . .
#
loop_
_ma_qa_metric_global.ordinal_id
_ma_qa_metric_global.model_id
_ma_qa_metric_global.metric_id
_ma_qa_metric_global.metric_value
1 1 1 1.0
2 1 2 2.0
3 1 3 3.0
4 1 4 4.0
"""
        s, = modelcif.reader.read(StringIO(cif))
        mg, = s.model_groups
        m, = mg
        q1, q2, q3, q4 = m.qa_metrics
        self.assertIsInstance(q1, modelcif.qa_metric.Global)
        self.assertEqual(q1.type, "other")
        self.assertEqual(q1.name, "MPQS")
        self.assertEqual(type(q1).__name__, "MPQS")
        self.assertEqual(q1.description, "ModPipe Quality Score")
        self.assertEqual(q1.__doc__, "ModPipe Quality Score")
        self.assertEqual(q1.software._id, '1')
        self.assertAlmostEqual(q1.value, 1.0, delta=1e-6)

        self.assertIsInstance(q2, modelcif.qa_metric.Global)
        self.assertIsInstance(q2, modelcif.qa_metric.ZScore)
        self.assertAlmostEqual(q2.value, 2.0, delta=1e-6)

        self.assertIsInstance(q3, modelcif.qa_metric.Global)
        self.assertIsInstance(q3, modelcif.qa_metric.Distance)
        self.assertAlmostEqual(q3.value, 3.0, delta=1e-6)

        self.assertIsInstance(q4, modelcif.qa_metric.Global)
        self.assertIsInstance(q4, modelcif.qa_metric.NormalizedScore)
        self.assertIsNone(q4.description)
        self.assertIsNone(q4.__doc__)

    def test_qa_metric_local_handler(self):
        """Test _QAMetricLocalHandler"""
        cif = """
loop_
_ma_model_list.ordinal_id
_ma_model_list.model_id
_ma_model_list.model_group_id
_ma_model_list.model_name
_ma_model_list.model_group_name
_ma_model_list.assembly_id
_ma_model_list.data_id
_ma_model_list.model_type
_ma_model_list.model_type_other_details
1 1 1 'Best scoring model' 'All models' 1 4 'Homology model' .
#
loop_
_ma_qa_metric.id
_ma_qa_metric.name
_ma_qa_metric.description
_ma_qa_metric.type
_ma_qa_metric.mode
_ma_qa_metric.type_other_details
_ma_qa_metric.software_group_id
1 'test local' 'some local score' 'normalized score' local . .
#
loop_
_ma_qa_metric_local.ordinal_id
_ma_qa_metric_local.model_id
_ma_qa_metric_local.label_asym_id
_ma_qa_metric_local.label_seq_id
_ma_qa_metric_local.label_comp_id
_ma_qa_metric_local.metric_id
_ma_qa_metric_local.metric_value
1 1 A 2 CYS 1 1.0
"""
        s, = modelcif.reader.read(StringIO(cif))
        mg, = s.model_groups
        m, = mg
        q1, = m.qa_metrics
        self.assertIsInstance(q1, modelcif.qa_metric.Local)
        self.assertIsInstance(q1, modelcif.qa_metric.NormalizedScore)
        self.assertEqual(q1.type, "normalized score")
        self.assertEqual(q1.name, "test local")
        self.assertEqual(q1.description, "some local score")
        self.assertIsNone(q1.software)
        self.assertEqual(q1.residue.asym._id, 'A')
        self.assertEqual(q1.residue.seq_id, 2)
        self.assertAlmostEqual(q1.value, 1.0, delta=1e-6)

    def test_qa_metric_pairwise_handler(self):
        """Test _QAMetricPairwiseHandler"""
        cif = """
loop_
_ma_model_list.ordinal_id
_ma_model_list.model_id
_ma_model_list.model_group_id
_ma_model_list.model_name
_ma_model_list.model_group_name
_ma_model_list.assembly_id
_ma_model_list.data_id
_ma_model_list.model_type
_ma_model_list.model_type_other_details
1 1 1 'Best scoring model' 'All models' 1 4 'Homology model' .
#
loop_
_ma_qa_metric.id
_ma_qa_metric.name
_ma_qa_metric.description
_ma_qa_metric.type
_ma_qa_metric.mode
_ma_qa_metric.type_other_details
_ma_qa_metric.software_group_id
1 'test pair' 'some pair score' 'normalized score' local-pairwise . .
#
loop_
_ma_qa_metric_local_pairwise.ordinal_id
_ma_qa_metric_local_pairwise.model_id
_ma_qa_metric_local_pairwise.label_asym_id_1
_ma_qa_metric_local_pairwise.label_seq_id_1
_ma_qa_metric_local_pairwise.label_comp_id_1
_ma_qa_metric_local_pairwise.label_asym_id_2
_ma_qa_metric_local_pairwise.label_seq_id_2
_ma_qa_metric_local_pairwise.label_comp_id_2
_ma_qa_metric_local_pairwise.metric_id
_ma_qa_metric_local_pairwise.metric_value
1 1 A 2 CYS B 4 GLY 1 1.0
"""
        s, = modelcif.reader.read(StringIO(cif))
        mg, = s.model_groups
        m, = mg
        q1, = m.qa_metrics
        self.assertIsInstance(q1, modelcif.qa_metric.LocalPairwise)
        self.assertIsInstance(q1, modelcif.qa_metric.NormalizedScore)
        self.assertEqual(q1.type, "normalized score")
        self.assertEqual(q1.name, "test pair")
        self.assertEqual(q1.description, "some pair score")
        self.assertIsNone(q1.software)
        self.assertEqual(q1.residue1.asym._id, 'A')
        self.assertEqual(q1.residue1.seq_id, 2)
        self.assertEqual(q1.residue2.asym._id, 'B')
        self.assertEqual(q1.residue2.seq_id, 4)
        self.assertAlmostEqual(q1.value, 1.0, delta=1e-6)

    def test_qa_metric_feature_handler(self):
        """Test _QAMetricFeatureHandler"""
        feat = """
loop_
_ma_atom_feature.ordinal_id
_ma_atom_feature.feature_id
_ma_atom_feature.atom_id
1 1 1
#
loop_
_ma_poly_residue_feature.ordinal_id
_ma_poly_residue_feature.feature_id
_ma_poly_residue_feature.label_asym_id
_ma_poly_residue_feature.label_seq_id
_ma_poly_residue_feature.label_comp_id
1 2 Y 1 ALA
#
loop_
_ma_entity_instance_feature.ordinal_id
_ma_entity_instance_feature.feature_id
_ma_entity_instance_feature.label_asym_id
1 3 Y
"""
        qa = """
loop_
_ma_feature_list.feature_id
_ma_feature_list.feature_type
_ma_feature_list.entity_type
_ma_feature_list.details
1 atom other 'atom f'
2 residue polymer prf
3 'entity instance' polymer .
#
loop_
_ma_model_list.ordinal_id
_ma_model_list.model_id
_ma_model_list.model_group_id
_ma_model_list.model_name
_ma_model_list.model_group_name
_ma_model_list.assembly_id
_ma_model_list.data_id
_ma_model_list.model_type
_ma_model_list.model_type_other_details
1 1 1 'Best scoring model' 'All models' 1 4 'Homology model' .
#
loop_
_ma_qa_metric.id
_ma_qa_metric.name
_ma_qa_metric.description
_ma_qa_metric.type
_ma_qa_metric.mode
_ma_qa_metric.type_other_details
_ma_qa_metric.software_group_id
1 'test local' 'some local score' 'normalized score' per-feature . .
#
loop_
_ma_qa_metric_feature.ordinal_id
_ma_qa_metric_feature.model_id
_ma_qa_metric_feature.feature_id
_ma_qa_metric_feature.metric_id
_ma_qa_metric_feature.metric_value
1 1 1 1 1.0
2 1 2 1 2.0
3 1 3 1 3.0
"""
        # Test both ways to make sure features still work if they are
        # referenced by ID before their type is known
        for cif in (feat + qa, qa + feat):
            s, = modelcif.reader.read(StringIO(cif))
            mg, = s.model_groups
            m, = mg
            q1, q2, q3 = m.qa_metrics
            self.assertIsInstance(q1, modelcif.qa_metric.Feature)
            self.assertIsInstance(q1, modelcif.qa_metric.NormalizedScore)
            self.assertIsInstance(q1.feature, modelcif.AtomFeature)
            self.assertEqual(q1.feature.details, 'atom f')
            self.assertAlmostEqual(q1.value, 1.0, delta=1e-6)
            self.assertIsInstance(q2.feature, modelcif.PolyResidueFeature)
            self.assertEqual(len(q2.feature.residues), 1)
            self.assertEqual(q2.feature.residues[0].seq_id, 1)
            self.assertAlmostEqual(q2.value, 2.0, delta=1e-6)
            self.assertIsInstance(q3.feature, modelcif.EntityInstanceFeature)
            self.assertEqual(len(q3.feature.asym_units), 1)
            self.assertAlmostEqual(q3.value, 3.0, delta=1e-6)

    def test_qa_metric_feature_pairwise_handler(self):
        """Test _QAMetricFeaturePairwiseHandler"""
        feat = """
loop_
_ma_poly_residue_feature.ordinal_id
_ma_poly_residue_feature.feature_id
_ma_poly_residue_feature.label_asym_id
_ma_poly_residue_feature.label_seq_id
_ma_poly_residue_feature.label_comp_id
1 1 Y 1 ALA
2 2 Y 2 CYS
"""
        qa = """
loop_
_ma_feature_list.feature_id
_ma_feature_list.feature_type
_ma_feature_list.entity_type
_ma_feature_list.details
1 residue polymer .
2 residue polymer .
#
loop_
_ma_model_list.ordinal_id
_ma_model_list.model_id
_ma_model_list.model_group_id
_ma_model_list.model_name
_ma_model_list.model_group_name
_ma_model_list.assembly_id
_ma_model_list.data_id
_ma_model_list.model_type
_ma_model_list.model_type_other_details
1 1 1 'Best scoring model' 'All models' 1 4 'Homology model' .
#
loop_
_ma_qa_metric.id
_ma_qa_metric.name
_ma_qa_metric.description
_ma_qa_metric.type
_ma_qa_metric.mode
_ma_qa_metric.type_other_details
_ma_qa_metric.software_group_id
1 'test local' 'some local score' 'normalized score' per-feature-pair . .
#
loop_
_ma_qa_metric_feature_pairwise.ordinal_id
_ma_qa_metric_feature_pairwise.model_id
_ma_qa_metric_feature_pairwise.feature_id_1
_ma_qa_metric_feature_pairwise.feature_id_2
_ma_qa_metric_feature_pairwise.metric_id
_ma_qa_metric_feature_pairwise.metric_value
1 1 1 2 1 50.000
"""
        # Test both ways to make sure features still work if they are
        # referenced by ID before their type is known
        for cif in (feat + qa, qa + feat):
            s, = modelcif.reader.read(StringIO(cif))
            mg, = s.model_groups
            m, = mg
            q1, = m.qa_metrics
            self.assertIsInstance(q1, modelcif.qa_metric.FeaturePairwise)
            self.assertIsInstance(q1, modelcif.qa_metric.NormalizedScore)
            self.assertIsInstance(q1.feature1, modelcif.PolyResidueFeature)
            self.assertIsInstance(q1.feature2, modelcif.PolyResidueFeature)
            self.assertAlmostEqual(q1.value, 50.0, delta=1e-6)

    def test_qa_metric_dihedral_handler(self):
        """Test _QAMetricDihedralHandler"""
        atoms = """
loop_
_atom_site.group_PDB
_atom_site.id
_atom_site.type_symbol
_atom_site.label_atom_id
_atom_site.label_alt_id
_atom_site.label_comp_id
_atom_site.label_seq_id
_atom_site.auth_seq_id
_atom_site.pdbx_PDB_ins_code
_atom_site.label_asym_id
_atom_site.Cartn_x
_atom_site.Cartn_y
_atom_site.Cartn_z
_atom_site.occupancy
_atom_site.label_entity_id
_atom_site.auth_asym_id
_atom_site.B_iso_or_equiv
_atom_site.pdbx_PDB_model_num
ATOM 1000 C CA . ASP 1 1 ? A 1.000 2.000 3.000 . 1 A . 1
ATOM 1900 C CA . ASP 1 1 ? A 1.000 2.000 3.000 . 1 A . 1
ATOM 2000 C CA . ASP 1 1 ? A 1.000 2.000 3.000 . 1 A . 2
ATOM 2900 C CA . ASP 1 1 ? A 1.000 2.000 3.000 . 1 A . 2
"""
        qa = """
loop_
_ma_qa_metric.id
_ma_qa_metric.name
_ma_qa_metric.description
_ma_qa_metric.type
_ma_qa_metric.mode
_ma_qa_metric.type_other_details
_ma_qa_metric.software_group_id
1 'test dihedral' 'some dihedral score' 'energy' dihedral . .
#
loop_
_ma_qa_metric_dihedral.ordinal_id
_ma_qa_metric_dihedral.atom_id_1
_ma_qa_metric_dihedral.atom_id_2
_ma_qa_metric_dihedral.atom_id_3
_ma_qa_metric_dihedral.atom_id_4
_ma_qa_metric_dihedral.metric_id
_ma_qa_metric_dihedral.metric_value
_ma_qa_metric_dihedral.quality
_ma_qa_metric_dihedral.smarts_pattern
1 1407 1410 1403 1404 1 10.0 TOLERABLE .
2 2407 2410 2403 2404 1 20.0 invalid-quality some-smarts
"""
        # Test both ways to make sure dihedrals still work if they reference
        # atoms by ID before we read the atom_site table
        for cif in (atoms + qa, qa + atoms):
            s, = modelcif.reader.read(StringIO(cif))
            mg, = s.model_groups
            m1, m2 = mg
            q1, = m1.qa_metrics
            q2, = m2.qa_metrics
            self.assertIsInstance(q1, modelcif.qa_metric.Dihedral)
            self.assertIsInstance(q1, modelcif.qa_metric.Energy)
            self.assertAlmostEqual(q1.value, 10.0, delta=1e-6)
            self.assertEqual(q1.quality, "tolerable")
            self.assertIsNone(q1.smarts_pattern)

            self.assertAlmostEqual(q2.value, 20.0, delta=1e-6)
            self.assertIsNone(q2.quality)
            self.assertEqual(q2.smarts_pattern, "some-smarts")

    def test_alignment_info_details_handler(self):
        """Test _AlignmentInfoHandler and _AlignmentDetailsHandler"""
        cif = """
loop_
_ma_alignment_info.alignment_id
_ma_alignment_info.data_id
_ma_alignment_info.software_group_id
_ma_alignment_info.alignment_length
_ma_alignment_info.alignment_type
_ma_alignment_info.alignment_mode
1 3 1 . 'target-template pairwise alignment' global
2 4 1 . 'target-template pairwise alignment' global
3 5 1 . 'target-template MSA' local
#
#
loop_
_ma_alignment_details.ordinal_id
_ma_alignment_details.alignment_id
_ma_alignment_details.template_segment_id
_ma_alignment_details.target_asym_id
_ma_alignment_details.score_type
_ma_alignment_details.score_type_other_details
_ma_alignment_details.score_value
_ma_alignment_details.percent_sequence_identity
_ma_alignment_details.sequence_identity_denominator
_ma_alignment_details.sequence_identity_denominator_other_details
1 1 1 A 'BLAST e-value' . 1.0 45.000 'Length of the shorter sequence' .
2 2 1 A . . . . . .
3 3 1 A 'HHblits e-value' . 2.0 45.000 'Arithmetic mean sequence length' .
#
loop_
_ma_alignment.ordinal_id
_ma_alignment.alignment_id
_ma_alignment.target_template_flag
_ma_alignment.sequence
1 1 1 DSYV-ETLD
2 1 2 DMACDTFIK
3 1 1 DSYV-ETLD
4 1 2 DMACDTFIK
#
loop_
_ma_target_template_poly_mapping.id
_ma_target_template_poly_mapping.template_segment_id
_ma_target_template_poly_mapping.target_asym_id
_ma_target_template_poly_mapping.target_seq_id_begin
_ma_target_template_poly_mapping.target_seq_id_end
1 1 A 1 8
2 1 A 1 8
"""
        s, = modelcif.reader.read(StringIO(cif))
        a1, a2, a3, = s.alignments
        self.assertIs(a1.__class__, a2.__class__)
        self.assertIsInstance(a1, modelcif.alignment.Global)
        self.assertIsInstance(a1, modelcif.alignment.Pairwise)
        p, = a1.pairs
        self.assertIsInstance(p.score, modelcif.alignment.BLASTEValue)
        self.assertAlmostEqual(p.score.value, 1.0, delta=1e-6)
        self.assertIsInstance(p.identity,
                              modelcif.alignment.ShorterSequenceIdentity)
        self.assertAlmostEqual(p.identity.value, 45.0, delta=1e-6)
        self.assertIsInstance(p.template, modelcif.TemplateSegment)
        self.assertEqual(p.template._id, '1')
        self.assertEqual(p.template.gapped_sequence, 'DMACDTFIK')
        self.assertIsInstance(p.target, ihm.AsymUnitSegment)
        self.assertEqual(p.target.asym._id, 'A')
        self.assertEqual(p.target.gapped_sequence, 'DSYV-ETLD')
        self.assertEqual(p.target.seq_id_range, (1, 8))
        self.assertIsInstance(a3, modelcif.alignment.Local)
        self.assertIsInstance(a3, modelcif.alignment.Multiple)
        p, = a2.pairs
        self.assertIsNone(p.score)
        self.assertIsNone(p.identity)
        p, = a3.pairs
        self.assertIsInstance(p.score, modelcif.alignment.HHblitsEValue)
        self.assertAlmostEqual(p.score.value, 2.0, delta=1e-6)
        self.assertIsInstance(p.identity,
                              modelcif.alignment.MeanSequenceIdentity)

    def test_associated_files(self):
        """Test _AssociatedHandler and _AssociatedArchiveHandler"""
        cif = """
loop_
_ma_data.id
_ma_data.name
_ma_data.content_type
_ma_data.content_type_other_details
42 'Model subunit' target .
loop_
_ma_target_entity.entity_id
_ma_target_entity.data_id
_ma_target_entity.origin
1 99 'reference database'
loop_
_ma_entry_associated_files.id
_ma_entry_associated_files.entry_id
_ma_entry_associated_files.file_url
_ma_entry_associated_files.file_type
_ma_entry_associated_files.file_format
_ma_entry_associated_files.file_content
_ma_entry_associated_files.details
_ma_entry_associated_files.data_id
1 model https://example.com/foo.txt file other other 'test file' .
2 model https://example.com/t.zip archive zip 'archive with multiple files' . .
3 model baz.txt file other other 'test file3' .
4 model baz.cif file cif other 'test mmCIF' .
5 model baz.bcif file bcif other 'test BinaryCIF' 42
#
#
loop_
_ma_associated_archive_file_details.id
_ma_associated_archive_file_details.archive_file_id
_ma_associated_archive_file_details.file_path
_ma_associated_archive_file_details.file_format
_ma_associated_archive_file_details.file_content
_ma_associated_archive_file_details.description
_ma_associated_archive_file_details.data_id
1 2 bar.txt other other 'test file2' .
2 99 99.txt other other 'test file99' .
3 2 bar.cif cif other 'test mmCIF in zip' .
4 2 bar.bcif bcif 'local pairwise QA scores' 'test BinaryCIF in zip' 99
5 2 bar2.bcif bcif 'QA metrics' 'test BinaryCIF in zip' 99
"""
        s, = modelcif.reader.read(StringIO(cif))
        r1, r2 = s.repositories
        self.assertEqual(r1.url_root, 'https://example.com')
        f1, zf = r1.files
        self.assertIsInstance(f1, modelcif.associated.File)
        self.assertEqual(f1.path, 'foo.txt')
        self.assertEqual(f1.details, 'test file')

        self.assertIsInstance(zf, modelcif.associated.ZipFile)
        self.assertEqual(zf.path, 't.zip')
        self.assertIsNone(zf.details)

        f2, f3, f4, f5 = zf.files
        self.assertEqual(f2.path, 'bar.txt')
        self.assertEqual(f2.details, 'test file2')
        self.assertIsNone(f2.data)
        self.assertIsInstance(f3, modelcif.associated.CIFFile)
        self.assertFalse(f3.binary)
        # QA metrics file using old "local pairwise QA scores" name
        self.assertIsInstance(
            f4, modelcif.associated.QAMetricsFile)
        self.assertEqual(f4.file_content, 'QA metrics')
        self.assertTrue(f4.binary)
        self.assertIsInstance(f4.data, modelcif.Entity)
        self.assertIsInstance(
            f5, modelcif.associated.QAMetricsFile)
        self.assertEqual(f5.file_content, 'QA metrics')
        self.assertTrue(f5.binary)
        self.assertIsInstance(f5.data, modelcif.Entity)

        self.assertIsNone(r2.url_root)
        f3, f4, f5 = r2.files
        self.assertEqual(f3.path, 'baz.txt')
        self.assertEqual(f3.details, 'test file3')
        self.assertIsInstance(f4, modelcif.associated.CIFFile)
        self.assertFalse(f4.binary)
        self.assertIsNone(f4.data)
        self.assertIsInstance(f5, modelcif.associated.CIFFile)
        self.assertTrue(f5.binary)
        self.assertEqual(f5.data.__class__, modelcif.data.Data)

    def test_template_poly_handler(self):
        """Test _TemplatePolyHandler"""
        cif = """
loop_
_chem_comp.id
_chem_comp.type
_chem_comp.name
_chem_comp.formula
MYTYPE 'D-PEPTIDE LINKING' 'MY CUSTOM COMPONENT' 'C6 H12'
MYTYP2 'D-PEPTIDE LINKING' 'MY CUSTOM COMPONENT2' 'C6 H12'
#
loop_
_ma_template_details.ordinal_id
_ma_template_details.template_id
_ma_template_details.template_origin
_ma_template_details.template_entity_type
_ma_template_details.template_trans_matrix_id
_ma_template_details.template_data_id
_ma_template_details.target_asym_id
_ma_template_details.template_label_asym_id
_ma_template_details.template_label_entity_id
_ma_template_details.template_model_num
_ma_template_details.template_auth_asym_id
1 1 'reference database' polymer 1 1 A A . 1 A
2 2 'reference database' polymer 1 1 A A . 1 A
3 3 'reference database' polymer 1 1 A A . 1 A
#
loop_
_ma_template_poly.template_id
_ma_template_poly.seq_one_letter_code
_ma_template_poly.seq_one_letter_code_can
1 A(MYTYPE)V AVV
2 A(MYTYP2)V .
3 . .
4 CCC CCC
"""
        s, = modelcif.reader.read(StringIO(cif))
        # template_id=4 in template_poly should be ignored
        t1, t2, t3 = s.templates
        s1, s2, s3 = t1.entity.sequence
        self.assertEqual(s1.id, 'ALA')
        self.assertEqual(s1.code, 'A')
        # Both one-letter and one-letter-canonical were provided
        self.assertEqual(s2.id, 'MYTYPE')
        self.assertEqual(s2.code, 'MYTYPE')
        self.assertEqual(s2.code_canonical, 'V')

        # Only one-letter was provided
        s1, s2, s3 = t2.entity.sequence
        self.assertEqual(s2.id, 'MYTYP2')
        self.assertEqual(s2.code, 'MYTYP2')
        self.assertIsNone(s2.code_canonical)

        # No sequence provided
        self.assertEqual(len(t3.entity.sequence), 0)

    def test_template_non_poly_handler(self):
        """Test _TemplateNonPolyHandler"""
        cif = """
loop_
_chem_comp.id
_chem_comp.type
HEM non-polymer
#
loop_
_ma_template_details.ordinal_id
_ma_template_details.template_id
_ma_template_details.template_origin
_ma_template_details.template_entity_type
_ma_template_details.template_trans_matrix_id
_ma_template_details.template_data_id
_ma_template_details.target_asym_id
_ma_template_details.template_label_asym_id
_ma_template_details.template_label_entity_id
_ma_template_details.template_model_num
_ma_template_details.template_auth_asym_id
1 1 'reference database' polymer 1 1 A A . 1 A
#
loop_
_ma_template_non_poly.template_id
_ma_template_non_poly.comp_id
_ma_template_non_poly.details
1 HEM Heme
"""
        s, = modelcif.reader.read(StringIO(cif))
        t, = s.templates
        s1, = t.entity.sequence
        self.assertEqual(s1.id, 'HEM')
        self.assertEqual(s1.type, 'non-polymer')
        self.assertIsInstance(s1, ihm.NonPolymerChemComp)

    def test_chem_comp_handler(self):
        """Test ChemCompHandler and ChemCompDescriptorHandler"""
        cif = """
loop_
_chem_comp.id
_chem_comp.type
_chem_comp.name
_chem_comp.formula
_chem_comp.ma_provenance
MET 'L-peptide linking' . . .
CYS 'D-peptide linking' CYSTEINE . ?
ALA 'L-peptide linking' ALANINE . 'CCD Core'
MATYPE 'L-PEPTIDE LINKING' 'MODELARCHIVE COMPONENT' . 'CCD MA'
MYTYPE 'L-PEPTIDE LINKING' 'MY CUSTOM COMPONENT' . 'CCD local'
#
loop_
_ma_chem_comp_descriptor.ordinal_id
_ma_chem_comp_descriptor.chem_comp_id
_ma_chem_comp_descriptor.chem_comp_name
_ma_chem_comp_descriptor.type
_ma_chem_comp_descriptor.value
_ma_chem_comp_descriptor.details
_ma_chem_comp_descriptor.software_id
1 MYTYPE 'ignored' 'InChI Key' XDAOLTSRNUSPPH-XMMPIXPASA-N foo 1
2 MYTYPE ? 'IUPAC Name' foobar . .
#
loop_
_entity_poly_seq.entity_id
_entity_poly_seq.num
_entity_poly_seq.mon_id
_entity_poly_seq.hetero
1 1 MET .
1 2 CYS .
1 3 ALA .
1 4 MATYPE .
1 5 MYTYPE .
"""
        s, = modelcif.reader.read(StringIO(cif))
        e1, = s.entities
        s = e1.sequence
        self.assertEqual(len(s), 5)
        self.assertEqual(s[2].ccd, 'core')
        self.assertEqual(s[3].ccd, 'ma')
        self.assertEqual(s[4].ccd, 'local')
        d1, d2 = s[4].descriptors
        self.assertIsInstance(d1, modelcif.descriptor.InChIKey)
        self.assertEqual(d1.value, 'XDAOLTSRNUSPPH-XMMPIXPASA-N')
        self.assertEqual(d1.details, 'foo')
        self.assertEqual(d1.software._id, '1')
        self.assertIsInstance(d2, modelcif.descriptor.IUPACName)
        self.assertEqual(d2.value, 'foobar')
        self.assertIsNone(d2.details)
        self.assertIsNone(d2.software)

    def test_add_to_system(self):
        """Test adding new mmCIF input to existing System"""
        s = modelcif.System()
        e = modelcif.Entity('AHC')
        e._id = '42'
        s.entities.append(e)
        fh = StringIO("""
loop_
_struct_asym.id
_struct_asym.entity_id
_struct_asym.details
A 42 foo
B 99 bar
""")
        s2, = modelcif.reader.read(fh, add_to_system=s)
        self.assertIs(s2, s)
        self.assertEqual(len(s.asym_units), 2)
        # asym A should point to existing entity
        self.assertEqual(s.asym_units[0].id, 'A')
        self.assertIs(s.asym_units[0].entity, e)

    def test_audit_revision_handler(self):
        """Test AuditRevisionHistoryHandler"""
        # We leverage the support in python-ihm, so only a basic test here
        cif = """
loop_
_pdbx_audit_revision_history.ordinal
_pdbx_audit_revision_history.data_content_type
_pdbx_audit_revision_history.major_revision
_pdbx_audit_revision_history.minor_revision
_pdbx_audit_revision_history.revision_date
40 'Structure model' 1 0 ?
41 'Structure model' 1 0 .
42 'Structure model' 2 0 1979-05-03
"""
        s, = modelcif.reader.read(StringIO(cif))
        r1, r2, r3 = s.revisions
        self.assertEqual(r3.major, 2)
        self.assertEqual(r3.minor, 0)
        self.assertEqual(r3.date, datetime.date(1979, 5, 3))

    def test_data_usage_handler(self):
        """Test DataUsageHandler"""
        # We leverage the support in python-ihm, so only a basic test here
        cif = """
loop_
_pdbx_data_usage.id
_pdbx_data_usage.type
_pdbx_data_usage.details
_pdbx_data_usage.url
_pdbx_data_usage.name
1 license 'some license' someurl somename
"""
        s, = modelcif.reader.read(StringIO(cif))
        d1, = s.data_usage
        self.assertEqual(d1.details, "some license")

    def test_atom_site_handler_water(self):
        """Test AtomSiteHandler reading water molecules"""
        fh = StringIO("""
loop_
_entity.id
_entity.type
1 water
loop_
_struct_asym.id
_struct_asym.entity_id
_struct_asym.details
A 1 Water
B 1 Water
#
loop_
_pdbx_nonpoly_scheme.asym_id
_pdbx_nonpoly_scheme.entity_id
_pdbx_nonpoly_scheme.mon_id
_pdbx_nonpoly_scheme.ndb_seq_num
_pdbx_nonpoly_scheme.pdb_seq_num
_pdbx_nonpoly_scheme.auth_seq_num
_pdbx_nonpoly_scheme.auth_mon_id
_pdbx_nonpoly_scheme.pdb_strand_id
_pdbx_nonpoly_scheme.pdb_ins_code
A 1 HOH 1 50 500 HOH A .
#
loop_
_atom_site.group_PDB
_atom_site.id
_atom_site.type_symbol
_atom_site.label_atom_id
_atom_site.label_alt_id
_atom_site.label_comp_id
_atom_site.label_seq_id
_atom_site.auth_seq_id
_atom_site.pdbx_PDB_ins_code
_atom_site.label_asym_id
_atom_site.Cartn_x
_atom_site.Cartn_y
_atom_site.Cartn_z
_atom_site.occupancy
_atom_site.label_entity_id
_atom_site.auth_asym_id
_atom_site.B_iso_or_equiv
_atom_site.pdbx_PDB_model_num
_atom_site.ihm_model_id
HETATM 1 O O . HOH . 40 ? A 10.000 10.000 10.000 . 1 A . 1 1
HETATM 2 O O . HOH . 50 ? A 10.000 10.000 10.000 . 1 A . 1 1
HETATM 3 O O . HOH . 60 . A 20.000 20.000 20.000 . 1 A . 1 1
HETATM 4 O O . HOH . 70 . B 20.000 20.000 20.000 . 1 B . 1 1
""")
        s, = modelcif.reader.read(fh)
        m = s.model_groups[0][0]
        a1, a2, a3, b1 = m._atoms
        # Should include info from both atom_site and scheme table
        self.assertEqual(a1.asym_unit.auth_seq_id_map,
                         {1: (40, None), 2: (50, None), 3: (60, None)})
        self.assertEqual(a1.asym_unit.orig_auth_seq_id_map,
                         {2: 500})
        self.assertEqual(b1.asym_unit.auth_seq_id_map, {1: (70, None)})
        self.assertIsNone(b1.asym_unit.orig_auth_seq_id_map)
        # Should get a WaterAsymUnit, not regular AsymUnit
        self.assertIsInstance(a1.asym_unit, modelcif.WaterAsymUnit)
        self.assertIsInstance(b1.asym_unit, modelcif.WaterAsymUnit)
        # seq_id should be assigned based on atom_site
        self.assertEqual(a1.seq_id, 1)
        self.assertEqual(a2.seq_id, 2)
        self.assertEqual(a3.seq_id, 3)
        self.assertEqual(b1.seq_id, 1)

    def test_nonpoly_scheme_handler(self):
        """Test NonPolySchemeHandler"""
        fh = StringIO("""
loop_
_chem_comp.id
_chem_comp.type
_chem_comp.name
CA non-polymer 'CALCIUM ION'
#
loop_
_entity.id
_entity.type
_entity.pdbx_description
1 non-polymer 'CALCIUM ION entity'
2 non-polymer 'no-chem-comp entity'
3 water       'no-chem-comp water'
#
loop_
_pdbx_entity_nonpoly.entity_id
_pdbx_entity_nonpoly.name
_pdbx_entity_nonpoly.comp_id
1 'CALCIUM ION'             CA
#
loop_
_struct_asym.id
_struct_asym.entity_id
_struct_asym.details
A 1 foo
B 2 bar
C 3 baz
#
loop_
_pdbx_nonpoly_scheme.asym_id
_pdbx_nonpoly_scheme.entity_id
_pdbx_nonpoly_scheme.mon_id
_pdbx_nonpoly_scheme.ndb_seq_num
_pdbx_nonpoly_scheme.pdb_seq_num
_pdbx_nonpoly_scheme.auth_seq_num
_pdbx_nonpoly_scheme.pdb_strand_id
_pdbx_nonpoly_scheme.pdb_ins_code
A 1 BAR 1 101 202 . .
B 2 BAR 1 1 1 Q X
C 3 HOH . 1 1 . .
C 3 HOH 2 2 2 . .
C 3 HOH 3 5 10 . .
C 3 HOH 4 1 20 . .
C 3 HOH 5 7 7 . .
""")
        s, = modelcif.reader.read(fh)
        e1, e2, e3 = s.entities
        # e1 should have sequence filled in by pdbx_entity_nonpoly
        self.assertEqual([cc.name for cc in e1.sequence], ['CALCIUM ION'])
        # e2,e3 should have sequence filled in by pdbx_nonpoly_scheme
        self.assertEqual([(cc.id, cc.name) for cc in e2.sequence],
                         [('BAR', 'no-chem-comp entity')])
        self.assertEqual([(cc.id, cc.name) for cc in e3.sequence],
                         [('HOH', 'WATER')])
        asym, a2, a3 = s.asym_units
        # non-polymers have no seq_id_range
        self.assertEqual(asym.seq_id_range, (None, None))
        self.assertEqual(asym.auth_seq_id_map, {1: (101, None)})
        self.assertEqual(asym.residue(1).auth_seq_id, 101)
        self.assertIsNone(asym.residue(1).ins_code)
        self.assertEqual(asym.strand_id, asym._id)
        self.assertIsNone(asym._strand_id)
        self.assertEqual(asym.orig_auth_seq_id_map, {1: 202})

        self.assertEqual(a2.auth_seq_id_map, {1: (1, 'X')})
        self.assertEqual(a2.residue(1).auth_seq_id, 1)
        self.assertEqual(a2.residue(1).ins_code, 'X')
        self.assertEqual(a2.strand_id, 'Q')
        self.assertEqual(a2._strand_id, 'Q')
        self.assertIsNone(a2.orig_auth_seq_id_map)

        self.assertEqual(a3.auth_seq_id_map, {1: (1, None), 2: (2, None),
                                              3: (5, None), 4: (1, None),
                                              5: (7, None)})
        self.assertEqual(a3.orig_auth_seq_id_map, {3: 10, 4: 20})

    def test_poly_seq_scheme_handler_offset(self):
        """Test PolySeqSchemeHandler with constant offset"""
        fh = StringIO(ASYM_ENTITY + """
loop_
_pdbx_poly_seq_scheme.asym_id
_pdbx_poly_seq_scheme.entity_id
_pdbx_poly_seq_scheme.seq_id
_pdbx_poly_seq_scheme.pdb_seq_num
_pdbx_poly_seq_scheme.pdb_strand_id
A 1 1 6 A
A 1 2 7 A
A 1 3 8 A
A 1 4 9 A
""")
        s, = modelcif.reader.read(fh)
        asym, = s.asym_units
        self.assertEqual(asym.auth_seq_id_map, 5)
        self.assertIsNone(asym._strand_id)
        self.assertEqual([asym.residue(i).auth_seq_id for i in range(1, 5)],
                         [6, 7, 8, 9])
        self.assertIsNone(asym.orig_auth_seq_id_map)

    def test_poly_seq_scheme_handler_offset_ins_code(self):
        """Test PolySeqSchemeHandler with constant offset but inscodes"""
        fh = StringIO(ASYM_ENTITY + """
loop_
_pdbx_poly_seq_scheme.asym_id
_pdbx_poly_seq_scheme.entity_id
_pdbx_poly_seq_scheme.seq_id
_pdbx_poly_seq_scheme.pdb_seq_num
_pdbx_poly_seq_scheme.pdb_strand_id
_pdbx_poly_seq_scheme.pdb_ins_code
A 1 1 6 A .
A 1 2 7 A .
A 1 3 8 A .
A 1 4 9 A A
""")
        s, = modelcif.reader.read(fh)
        asym, = s.asym_units
        self.assertEqual(asym.auth_seq_id_map,
                         {1: (6, None), 2: (7, None), 3: (8, None),
                          4: (9, 'A')})
        self.assertIsNone(asym._strand_id)
        self.assertEqual([asym.residue(i).auth_seq_id for i in range(1, 5)],
                         [6, 7, 8, 9])
        self.assertIsNone(asym.residue(1).ins_code)
        self.assertEqual(asym.residue(4).ins_code, 'A')
        self.assertIsNone(asym.orig_auth_seq_id_map)

    def test_poly_seq_scheme_handler_empty(self):
        """Test PolySeqSchemeHandler with no poly_seq_scheme"""
        fh = StringIO(ASYM_ENTITY)
        s, = modelcif.reader.read(fh)
        asym, = s.asym_units
        self.assertEqual(asym.auth_seq_id_map, 0)
        self.assertEqual([asym.residue(i).auth_seq_id for i in range(1, 5)],
                         [1, 2, 3, 4])
        self.assertIsNone(asym.orig_auth_seq_id_map)

    def test_poly_seq_scheme_handler_nop(self):
        """Test PolySeqSchemeHandler with a do-nothing poly_seq_scheme"""
        fh = StringIO(ASYM_ENTITY + """
loop_
_pdbx_poly_seq_scheme.asym_id
_pdbx_poly_seq_scheme.entity_id
_pdbx_poly_seq_scheme.seq_id
_pdbx_poly_seq_scheme.pdb_seq_num
A 1 1 1
A 1 2 2
A 1 3 3
""")
        s, = modelcif.reader.read(fh)
        asym, = s.asym_units
        self.assertEqual(asym.auth_seq_id_map, 0)
        self.assertEqual([asym.residue(i).auth_seq_id for i in range(1, 5)],
                         [1, 2, 3, 4])
        self.assertIsNone(asym.orig_auth_seq_id_map)

    def test_poly_seq_scheme_handler_partial(self):
        """Test PolySeqSchemeHandler with partial information"""
        fh = StringIO(ASYM_ENTITY + """
loop_
_pdbx_poly_seq_scheme.asym_id
_pdbx_poly_seq_scheme.entity_id
_pdbx_poly_seq_scheme.seq_id
_pdbx_poly_seq_scheme.pdb_seq_num
_pdbx_poly_seq_scheme.auth_seq_num
A 1 1 6 .
A 1 2 7 9
A 1 3 8 .
""")
        s, = modelcif.reader.read(fh)
        asym, = s.asym_units
        # No mapping for residue 4 (and no insertion codes at all)
        self.assertEqual(asym.auth_seq_id_map, {1: (6, None), 2: (7, None),
                                                3: (8, None)})
        self.assertEqual([asym.residue(i).auth_seq_id for i in range(1, 5)],
                         [6, 7, 8, 4])
        self.assertIsNone(asym.residue(1).ins_code)
        self.assertEqual(asym.orig_auth_seq_id_map, {2: 9})

    def test_poly_seq_scheme_handler_incon_off(self):
        """Test PolySeqSchemeHandler with inconsistent offset"""
        fh = StringIO(ASYM_ENTITY + """
loop_
_pdbx_poly_seq_scheme.asym_id
_pdbx_poly_seq_scheme.entity_id
_pdbx_poly_seq_scheme.seq_id
_pdbx_poly_seq_scheme.pdb_seq_num
_pdbx_poly_seq_scheme.pdb_strand_id
A 1 1 6 X
A 1 2 7 X
A 1 3 8 X
A 1 4 10 X
""")
        s, = modelcif.reader.read(fh)
        asym, = s.asym_units
        self.assertEqual(asym._strand_id, 'X')
        self.assertEqual(asym.auth_seq_id_map, {1: (6, None), 2: (7, None),
                                                3: (8, None), 4: (10, None)})
        self.assertEqual([asym.residue(i).auth_seq_id for i in range(1, 5)],
                         [6, 7, 8, 10])
        self.assertIsNone(asym.residue(1).ins_code)
        self.assertIsNone(asym.orig_auth_seq_id_map)

    def test_poly_seq_scheme_handler_unknown_auth_seq(self):
        """Test PolySeqSchemeHandler with explicit unknown auth_seq_num"""
        fh = StringIO(ASYM_ENTITY + """
loop_
_pdbx_poly_seq_scheme.asym_id
_pdbx_poly_seq_scheme.entity_id
_pdbx_poly_seq_scheme.seq_id
_pdbx_poly_seq_scheme.pdb_seq_num
_pdbx_poly_seq_scheme.auth_seq_num
_pdbx_poly_seq_scheme.pdb_strand_id
A 1 1 1 1 A
A 1 2 2 2 A
A 1 3 3 ? A
A 1 4 4 4 A
""")
        s, = modelcif.reader.read(fh)
        asym, = s.asym_units
        self.assertEqual(asym.auth_seq_id_map, 0)
        self.assertEqual(asym.orig_auth_seq_id_map, {3: ihm.unknown})

    def test_poly_seq_scheme_handler_str_seq_id(self):
        """Test PolySeqSchemeHandler with a non-integer pdb_seq_num"""
        fh = StringIO(ASYM_ENTITY + """
loop_
_pdbx_poly_seq_scheme.asym_id
_pdbx_poly_seq_scheme.entity_id
_pdbx_poly_seq_scheme.seq_id
_pdbx_poly_seq_scheme.pdb_seq_num
_pdbx_poly_seq_scheme.auth_seq_num
_pdbx_poly_seq_scheme.pdb_strand_id
_pdbx_poly_seq_scheme.pdb_ins_code
A 1 1 6 6 ? .
A 1 2 7 12 ? .
A 1 3 8 24 ? .
A 1 4 9A 48A ? .
""")
        s, = modelcif.reader.read(fh)
        asym, = s.asym_units
        self.assertIsNone(asym._strand_id)
        self.assertEqual(asym.auth_seq_id_map, {1: (6, None), 2: (7, None),
                                                3: (8, None), 4: ('9A', None)})
        self.assertEqual([asym.residue(i).auth_seq_id for i in range(1, 5)],
                         [6, 7, 8, '9A'])
        self.assertIsNone(asym.residue(1).ins_code)
        self.assertIsNone(asym.residue(3).ins_code)
        self.assertEqual(asym.orig_auth_seq_id_map, {2: 12, 3: 24, 4: '48A'})

    def test_branched(self):
        """Test read of branched entities"""
        # Tests for individual dumpers are already present in python-ihm;
        # here we just test to make sure they are being called by read()
        incif = utils.get_input_file_name(TOPDIR, 'mini_branched.cif')
        with open(incif) as fh:
            s, = modelcif.reader.read(fh)
        e, = s.entities
        a, = s.asym_units
        self.assertEqual(a.auth_seq_id_map,
                         {1: (51, None), 2: (52, None), 3: (53, None)})
        bd1, = e.branch_descriptors
        self.assertEqual(bd1.text, 'bar')
        lnk1, = e.branch_links
        self.assertEqual(lnk1.atom_id1, 'CA')


if __name__ == '__main__':
    unittest.main()



"""Classes to describe the chemistry of custom chemical components.

   If a given :class:`ihm.ChemComp` is not defined in either the wwPDB
   chemical component dictionary (CCD) or the ModelArchive CCD, its
   chemistry can be described with one or more of these objects. They
   are passed as the ``descriptors`` argument when creating a new
   :class:`ihm.ChemComp`.
"""


class Descriptor:
    """Base class for all descriptors.
       This class is generally not used directly; instead, a subclass
       such as :class:`IUPACName` or :class:`InChI` is employed.

       :param str value: The actual name or identifier describing
              the chemistry.
       :param str details: Additional details about this descriptor.
       :param software: The software used to generate the descriptor, if any.
       :type software: :class:`modelcif.Software`
    """

    def __init__(self, value, details=None, software=None):
        self.value, self.details = value, details
        self.software = software

    def __repr__(self):
        return "<%s(%s)>" % (self.__class__.__name__, repr(self.value))


class CanonicalSMILES(Descriptor):
    """Simplified Molecular-Input Line-Entry System (SMILES) computed from
       chemical structure devoid of isotopic and stereochemical information."""
    type = 'Canonical SMILES'


class IsomericSMILES(Descriptor):
    """Simplified Molecular-Input Line-Entry System (SMILES) computed from
       chemical structure containing isotopic and stereochemical information.

       SMILES written with isotopic and chiral specifications are collectively
       known as isomeric SMILES."""
    type = 'Isomeric SMILES'


class IUPACName(Descriptor):
    """Chemical name computed from chemical structure that uses International
       Union of Pure and Applied Chemistry (IUPAC) nomenclature standards."""
    type = 'IUPAC Name'


class InChI(Descriptor):
    """International Chemical Identifier (InChI) computed from chemical
       structure using the International Union of Pure and Applied Chemistry
       (IUPAC) standard."""
    type = 'InChI'


class InChIKey(Descriptor):
    """International Chemical Identifier hash (InChIKey) computed from
       chemical structure using the International Union of Pure and Applied
       Chemistry (IUPAC) standard."""
    type = 'InChI Key'


class PubChemCID(Descriptor):
    """PubChem Compound ID."""
    type = 'PubChem CID'



"""Classes to associate extra files with the system.

   Typically, one or more :class:`Repository` objects are created and
   added to :attr:`modelcif.System.repositories`."""

import posixpath
import warnings


class Repository:
    """An online location where associated files can be found.
       These objects are typically added to
       :attr:`modelcif.System.repositories`.

       :param str url_root: URL root that prefixes each file's path.
              For example, if url_root is ``https://example.com`` then
              a :class:`File` with path ``test.txt`` can be found at
              ``https://example.com/test.txt``. If the files are not
              available online, None can be used here.
       :param list files: A list of :class:`File` objects.
    """
    def __init__(self, url_root, files):
        self.url_root = url_root
        self.files = files

    def get_url(self, f):
        """Get the full URL for the given :class:`File`"""
        return posixpath.join(self.url_root or '', f.path)


class File:
    """A single associated file. These objects can be added to a
       :class:`Repository` or a :class:`ZipFile`.

       :param str path: File name.
       :param str details: Any additional information about the file.
       :param data: If available, the data (e.g. sequence, structure,
              alignment) that are stored in the file.
       :type data: :class:`~modelcif.data.Data`
    """
    file_type = 'file'
    file_content = 'other'
    file_format = 'other'

    def __init__(self, path, details=None, data=None):
        self.path, self.details, self.data = path, details, data


class CIFFile(File):
    """An associated file in mmCIF or BinaryCIF format.
       See :class:`File` for more details.

       :param str path: File name that will be used to construct URLs in the
              main mmCIF file (see :class:`Repository` or :class:`ZipFile`).
       :param str details: Any additional information about the file.
       :param data: If available, the data (e.g. sequence, structure,
              alignment) that are stored in the file.
       :type data: :class:`~modelcif.data.Data`
       :param list categories: If given, any mmCIF category names in this list
              are written out to ``local_path`` by
              :func:`modelcif.dumper.write` instead of to the primary file
              handle.
       :param list copy_categories: If given, any mmCIF category names in this
              list are written out to both ``local_path`` by
              :func:`modelcif.dumper.write` and the primary file handle.
       :param str entry_id: Unique identifier for the associated file,
              if written (by specifying ``categories`` or ``copy_categories``).
       :param str entry_details: A comment to be added to the associated file,
              if written (by specifying ``categories`` or ``copy_categories``).
       :param str local_path: File name that will be used for ``categories``
              or ``copy_categories``. If not given, it defaults to the same
              as ``path``. (The file is always written directly to the local
              disk, even if this object is placed inside a :class:`ZipFile`.)
       :param bool binary: If False (the default), any output file is written
              in mmCIF format; if True, the file is written in BinaryCIF.
    """

    _binary_ff_map = {True: 'bcif', False: 'cif'}

    file_format = property(lambda self: self._binary_ff_map[self.binary],
                           doc="Format of the file (BinaryCIF or mmCIF)")

    def __init__(self, path, details=None, categories=[], copy_categories=[],
                 entry_id='model', entry_details=None, local_path=None,
                 binary=False, data=None):
        super().__init__(path, details, data)
        self.categories = categories
        self.copy_categories = copy_categories
        self.id = entry_id
        self.entry_details = entry_details
        self.local_path = local_path or path
        self.binary = binary


class QAMetricsFile(CIFFile):
    """An associated file in CIF format containing QA metrics.
       See :class:`CIFFile` for more details.
    """
    file_content = 'QA metrics'


# Map old class name to new equivalent
class LocalPairwiseQAScoresFile(QAMetricsFile):
    def __init__(self, *args, **keys):
        warnings.warn("LocalPairwiseQAScoresFile is deprecated. "
                      "Use QAMetricsFile instead.", stacklevel=2)
        super().__init__(*args, **keys)


class ZipFile(File):
    """An associated archive file in zip format, containing other files.
       See :class:`File` for more details.

       :param list files: A list of the :class:`File` objects contained
              within this archive. Note that an archive cannot contain another
              archive.
    """
    file_type = 'archive'
    file_content = 'archive with multiple files'
    file_format = 'zip'

    def __init__(self, path, details=None, files=[], data=None):
        super().__init__(path, details, data)
        self.files = files



# This example demonstrates the use of the Python IHM library's validator.
# A structure is downloaded from the ModBase database and checked against
# the ModelCIF dictionary for compliance. This validator can be used
# to perform basic integrity checking against any mmCIF dictionary.
# See also validate_mmcif.py for a simpler script to validate a
# user-provided mmCIF file.

import io
import ihm.dictionary
import urllib.request

# Read in the ModelCIF dictionary from wwPDB as a Dictionary object.
# Note that the ModelCIF dictionary also includes the PDBx dictionary,
# so we don't need to read that in separately
fh = urllib.request.urlopen(
    'https://mmcif.wwpdb.org/dictionaries/ascii/mmcif_ma.dic')
pdbx_mc = ihm.dictionary.read(fh)
fh.close()

# Validate a structure against PDBx+ModelCIF.
# A correct structure here should result in no output; an invalid structure
# will result in a ValidatorError Python exception.
# Here, a structure from ModBase (which should be valid) is used.
acc = 'P21812'
cif = urllib.request.urlopen('https://salilab.org/modbase/retrieve'
                             '?databaseID=%s&format=mmcif' % acc).read()

# The encoding for mmCIF files isn't strictly defined, so first try UTF-8
# and if that fails, strip out any non-ASCII characters. This ensures that
# we handle accented characters in string fields correctly.
try:
    fh = io.StringIO(cif.decode('utf-8'))
except UnicodeDecodeError:
    fh = io.StringIO(cif.decode('ascii', errors='ignore'))

pdbx_mc.validate(fh)



import os
import sys
import tempfile
import contextlib
import shutil


def set_search_paths(topdir):
    """Set search paths so that we can import Python modules"""
    os.environ['PYTHONPATH'] = topdir + os.pathsep \
        + os.environ.get('PYTHONPATH', '')
    sys.path.insert(0, topdir)


def get_input_file_name(topdir, fname):
    """Return full path to a test input file"""
    return os.path.join(topdir, 'test', 'input', fname)


@contextlib.contextmanager
def temporary_directory(dir=None):
    _tmpdir = tempfile.mkdtemp(dir=dir)
    yield _tmpdir
    shutil.rmtree(_tmpdir, ignore_errors=True)


if 'coverage' in sys.modules:
    import atexit
    # Collect coverage information from subprocesses
    __site_tmpdir = tempfile.mkdtemp()
    with open(os.path.join(__site_tmpdir, 'sitecustomize.py'), 'w') as fh:
        fh.write("""
import coverage
import atexit
import os

_cov = coverage.coverage(branch=True, data_suffix=True, auto_data=True,
                         data_file=os.path.join('%s', '.coverage'))
_cov.start()

def _coverage_cleanup(c):
    c.stop()
atexit.register(_coverage_cleanup, _cov)
""" % os.getcwd())

    os.environ['PYTHONPATH'] = __site_tmpdir + os.pathsep \
        + os.environ.get('PYTHONPATH', '')

    def __cleanup(d):
        shutil.rmtree(d, ignore_errors=True)
    atexit.register(__cleanup, __site_tmpdir)



import utils
import os
import unittest

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import modelcif.reference


class Tests(unittest.TestCase):
    def test_template_reference(self):
        """Test TemplateReference classes"""
        ref = modelcif.reference.PDB("1abc")
        self.assertEqual(ref.name, "PDB")
        self.assertIsNone(ref.other_details)

        # generic "other" reference
        ref = modelcif.reference.TemplateReference("1abc")
        self.assertEqual(ref.name, "Other")
        self.assertIsNone(ref.other_details)

        # custom "other" reference
        class CustomRef(modelcif.reference.TemplateReference):
            """foo
               bar"""

        ref = CustomRef("1abc")
        self.assertEqual(ref.name, "Other")
        self.assertEqual(ref.other_details, "foo")

    def test_target_reference(self):
        """Test TargetReference classes"""
        ref = modelcif.reference.UniProt("code", "acc", sequence='CC')
        self.assertEqual(ref.name, "UNP")
        self.assertIsNone(ref.other_details)

        # Reference with (deprecated) align begin, end
        self.assertWarns(UserWarning, modelcif.reference.UniProt,
                         "code", "acc", align_begin=1, align_end=10,
                         sequence='CC')

        # Reference without explicit sequence
        self.assertWarns(UserWarning, modelcif.reference.UniProt,
                         "code", "acc")

        # generic "other" reference
        ref = modelcif.reference.TargetReference("code", "acc", sequence='CC')
        self.assertEqual(ref.name, "Other")
        self.assertIsNone(ref.other_details)

        # custom "other" reference
        class CustomRef(modelcif.reference.TargetReference):
            """foo
               bar"""

        ref = CustomRef("code", "acc", sequence='CC')
        self.assertEqual(ref.name, "Other")
        self.assertEqual(ref.other_details, "foo")


if __name__ == '__main__':
    unittest.main()



"""Classes to track inputs/outputs of modeling protocols.

   See also :class:`modelcif.protocol.Step`.
"""


class Data:
    """Some part of the system that is input or output by part of
       the modeling protocol, and/or stored in a file.

       Usually a subclass is passed to :class:`modelcif.protocol.Step`
       to describe the input or output, or to :class:`modelcif.associated.File`
       to point to where the data are stored:

        - A database of possible template sequences/structures to construct
          or search (:class:`modelcif.ReferenceDatabase`)
        - A template structure (:class:`modelcif.Template`)
        - The sequence of the target (:class:`modelcif.Entity`)
        - A target-template alignment (:mod:`modelcif.alignment`)
        - Target structure coordinates (:class:`modelcif.model.Model`)

       However, this class can also be used directly to describe other kinds
       of input/output data.

       :param str name: A short name for the data.
       :param str details: A longer description of the data.
    """
    data_content_type = 'other'
    data_other_details = None

    def __init__(self, name, details=None):
        self.name = name
        self.data_other_details = details


class DataGroup(list):
    """A number of :class:`Data` objects that are grouped together.

       This class can be used to group together multiple :class:`Data`
       objects if a given modeling protocol step consumes or generates
       multiple pieces of data. See :class:`modelcif.protocol.Step`. It behaves
       like a regular Python list.
    """
    pass



"""Utility classes to dump out information in mmCIF or BinaryCIF format"""

from datetime import date
import itertools
import operator
import ihm.dumper
import ihm
from ihm import util
import ihm.format
import ihm.format_bcif
from ihm.dumper import Dumper, Variant, _prettyprint_seq, _get_transform
import modelcif.qa_metric
import modelcif.data


class _AuditConformDumper(Dumper):
    URL = ("https://raw.githubusercontent.com/ihmwg/ModelCIF/%s/dist/" +
           "mmcif_ma.dic")

    def dump(self, system, writer):
        with writer.category("_audit_conform") as lp:
            # Update to match the version of the ModelCIF dictionary
            # we support:
            lp.write(dict_name="mmcif_ma.dic", dict_version="1.4.9",
                     dict_location=self.URL % "fece26d")


class _EntryLinkDumper(Dumper):
    def dump(self, system, writer):
        with writer.loop("_entry_link", ["id", "entry_id", "details"]) as lp:
            lp.write(id=1, entry_id=system.id, details=system.entry_details)


class _DatabaseDumper(Dumper):
    def dump(self, system, writer):
        if system.database:
            with writer.category("_database_2") as lp:
                lp.write(database_id=system.database.id,
                         database_code=system.database.code)


class _ChemCompDumper(Dumper):
    # Similar to ihm.dumper._ChemCompDumper, but we need to also include
    # components referenced only by Templates, as their Entities are not
    # included in system.entities by default

    _prov_map = {'core': 'CCD Core', 'ma': 'CCD MA', 'local': 'CCD local'}

    def _get_entities(self, system):
        return itertools.chain(
            system.entities, (t.entity for t in system.templates))

    def _get_provenance(self, comp):
        ccd = comp.ccd
        if ccd is None:
            ccd = 'core'
            if comp.descriptors:
                ccd = 'local'
        val = self._prov_map.get(ccd)
        if not val:
            if not self._check:
                return ihm.unknown
            raise KeyError("Invalid ccd value %s for %s; can be %s, or None"
                           % (repr(comp.ccd), comp,
                              ", ".join(sorted(self._prov_map.keys()))))
        return val

    def dump(self, system, writer):
        comps = frozenset(
            comp for e in self._get_entities(system) for comp in e.sequence)

        with writer.loop("_chem_comp", ["id", "type", "name",
                                        "formula", "formula_weight",
                                        "ma_provenance"]) as lp:
            for comp in sorted(comps, key=operator.attrgetter('id')):
                lp.write(id=comp.id, type=comp.type, name=comp.name,
                         formula=comp.formula,
                         formula_weight=comp.formula_weight,
                         ma_provenance=self._get_provenance(comp))


class _ChemCompDescriptorDumper(Dumper):
    def _get_entities(self, system):
        return itertools.chain(
            system.entities, (t.entity for t in system.templates))

    def dump(self, system, writer):
        ordinal = itertools.count(1)
        comps = frozenset(
            comp for e in self._get_entities(system) for comp in e.sequence)

        with writer.loop("_ma_chem_comp_descriptor",
                         ["ordinal_id", "chem_comp_id", "chem_comp_name",
                          "type", "value", "details", "software_id"]) as lp:
            for comp in sorted(comps, key=operator.attrgetter('id')):
                if not hasattr(comp, 'descriptors') or not comp.descriptors:
                    continue
                for desc in comp.descriptors:
                    lp.write(ordinal_id=next(ordinal), chem_comp_id=comp.id,
                             chem_comp_name=comp.name, type=desc.type,
                             value=desc.value, details=desc.details,
                             software_id=desc.software._id
                             if desc.software else None)


class _TargetRefDBDumper(Dumper):
    def dump(self, system, writer):
        with writer.loop(
                "_ma_target_ref_db_details",
                ["target_entity_id", "db_name", "db_name_other_details",
                 "db_code", "db_accession", "seq_db_isoform",
                 "seq_db_align_begin", "seq_db_align_end",
                 "ncbi_taxonomy_id", "organism_scientific",
                 "seq_db_sequence_version_date",
                 "seq_db_sequence_checksum",
                 "is_primary"]) as lp:
            for e in system.entities:
                for r in e.references:
                    if r.align_begin is None:
                        db_begin = min(a.db_begin for a in r._get_alignments())
                    else:
                        db_begin = r.align_begin
                    if r.align_end is None:
                        aligns = [a for a in r._get_alignments()
                                  if a.db_end is not None]
                        if aligns:
                            db_end = max(a.db_end for a in aligns)
                        else:
                            db_end = len(r.sequence)
                    else:
                        db_end = r.align_end
                    lp.write(target_entity_id=e._id, db_name=r.name,
                             db_name_other_details=r.other_details,
                             db_code=r.code, db_accession=r.accession,
                             seq_db_isoform=r.isoform,
                             seq_db_align_begin=db_begin,
                             seq_db_align_end=db_end,
                             ncbi_taxonomy_id=r.ncbi_taxonomy_id,
                             organism_scientific=r.organism_scientific,
                             seq_db_sequence_version_date=date.isoformat(
                                 r.sequence_version_date)
                             if r.sequence_version_date else None,
                             seq_db_sequence_checksum=r.sequence_crc64,
                             is_primary=r.is_primary)


class _EntityNonPolyDumper(Dumper):
    def finalize(self, system):
        self._ma_model_mode_map = {}
        expmap = {True: 'explicit', False: 'implicit'}
        for a in system.asym_units:
            if isinstance(a, modelcif.NonPolymerFromTemplate):
                self._ma_model_mode_map[a.template.entity] = \
                    expmap.get(a.explicit)

    def dump(self, system, writer):
        with writer.loop("_pdbx_entity_nonpoly",
                         ["entity_id", "name", "comp_id",
                          "ma_model_mode"]) as lp:
            for entity in system.entities:
                if entity.is_polymeric() or entity.is_branched():
                    continue
                lp.write(entity_id=entity._id, name=entity.description,
                         comp_id=entity.sequence[0].id,
                         ma_model_mode=self._ma_model_mode_map.get(entity))


class _TargetEntityDumper(Dumper):
    def dump(self, system, writer):
        with writer.loop(
                "_ma_target_entity",
                ["entity_id", "data_id", "origin"]) as lp:
            for e in system.entities:
                lp.write(entity_id=e._id, data_id=e._data_id,
                         origin="reference database" if e.references
                         else "designed")

        with writer.loop(
                "_ma_target_entity_instance",
                ["asym_id", "entity_id", "details"]) as lp:
            for asym in system.asym_units:
                lp.write(asym_id=asym._id, entity_id=asym.entity._id,
                         details=asym.details)


class _SoftwareGroupDumper(Dumper):
    def finalize(self, system):
        # Map from id(list) to id
        self._param_group_id = {}
        self._param_groups = []
        for n, s in enumerate(system.software_groups):
            # Use _group_id rather than _id as the "group" might be a
            # singleton Software, which already has its own id
            s._group_id = n + 1
            if isinstance(s, modelcif.SoftwareGroup):
                for soft in s:
                    if (isinstance(soft, modelcif.SoftwareWithParameters)
                            and soft.parameters
                            and id(soft.parameters)
                            not in self._param_group_id):
                        self._param_groups.append(soft.parameters)
                        self._param_group_id[id(soft.parameters)] \
                            = len(self._param_groups)

    def dump(self, system, writer):
        self.dump_parameters(system, writer)
        self.dump_groups(system, writer)

    def dump_groups(self, system, writer):
        ordinal = itertools.count(1)
        with writer.loop(
                "_ma_software_group",
                ["ordinal_id", "group_id", "software_id",
                 "parameter_group_id"]) as lp:
            for g in system.software_groups:
                if isinstance(g, modelcif.Software):
                    # If a singleton Software, write a group containing one
                    # member
                    lp.write(ordinal_id=next(ordinal), group_id=g._group_id,
                             software_id=g._id)
                else:
                    for s in g:
                        param = None
                        if isinstance(s, modelcif.SoftwareWithParameters):
                            soft_id = s.software._id
                            if s.parameters:
                                param = self._param_group_id[id(s.parameters)]
                        else:
                            soft_id = s._id
                        lp.write(ordinal_id=next(ordinal),
                                 group_id=g._group_id, software_id=soft_id,
                                 parameter_group_id=param)

    def _handle_list(self, value):
        list_type_map = {int: 'integer-csv', float: 'float-csv'}
        types = frozenset(type(x) for x in value)
        if types == frozenset((int,)):
            data_type = list_type_map[int]
        elif types == frozenset((float,)) or types == frozenset((int, float)):
            # Treat mix of int and float as float
            data_type = list_type_map[float]
        else:
            raise TypeError("Only lists of ints or floats are supported")
        return data_type, ",".join(str(x) for x in value)

    def dump_parameters(self, system, writer):
        parameter_id = itertools.count(1)
        type_map = {int: "integer", float: "float", str: "string",
                    bool: "boolean"}
        with writer.loop(
                "_ma_software_parameter",
                ["parameter_id", "group_id", "data_type",
                 "name", "value", "description"]) as lp:
            for g in self._param_groups:
                group_id = self._param_group_id[id(g)]
                for p in g:
                    if isinstance(p.value, (list, tuple)):
                        data_type, value = self._handle_list(p.value)
                    else:
                        data_type = type_map.get(type(p.value), str)
                        value = p.value
                    lp.write(parameter_id=next(parameter_id),
                             group_id=group_id, data_type=data_type,
                             name=p.name, value=value,
                             description=p.description)


class _DataDumper(Dumper):
    def finalize(self, system):
        for n, d in enumerate(system.data):
            d._data_id = n + 1

    def dump(self, system, writer):
        with writer.loop(
                "_ma_data",
                ["id", "name", "content_type",
                 "content_type_other_details"]) as lp:
            for d in system.data:
                # ihm.Entity isn't a subclass of Data, so we need
                # to fill in missing attributes here
                if isinstance(d, ihm.Entity):
                    lp.write(id=d._data_id, name=d.description,
                             content_type="target",
                             content_type_other_details=None)
                else:
                    lp.write(id=d._data_id, name=d.name,
                             content_type=d.data_content_type,
                             content_type_other_details=d.data_other_details)


class _DataGroupDumper(Dumper):
    def finalize(self, system):
        for n, d in enumerate(system.data_groups):
            # Use _data_group_id rather than _id as the "group" might be a
            # singleton Data, which already has its own id
            d._data_group_id = n + 1

    def dump(self, system, writer):
        ordinal = itertools.count(1)
        with writer.loop(
                "_ma_data_group",
                ["ordinal_id", "group_id", "data_id"]) as lp:
            for g in system.data_groups:
                if isinstance(g, (modelcif.data.Data, ihm.Entity)):
                    # If a singleton Data (or ihm.Entity, which isn't a
                    # subclass of Data), write a group containing one member
                    lp.write(ordinal_id=next(ordinal),
                             group_id=g._data_group_id, data_id=g._data_id)
                else:
                    for d in g:
                        lp.write(ordinal_id=next(ordinal),
                                 group_id=g._data_group_id, data_id=d._data_id)


class _DataRefDBDumper(Dumper):
    def dump(self, system, writer):
        with writer.loop(
                "_ma_data_ref_db",
                ["data_id", "name", "location_url",
                 "version", "release_date"]) as lp:
            for d in system.data:
                if not isinstance(d, modelcif.ReferenceDatabase):
                    continue
                lp.write(data_id=d._data_id, name=d.name, location_url=d.url,
                         version=d.version,
                         release_date=date.isoformat(d.release_date)
                         if d.release_date else None)


class _TemplateTransformDumper(Dumper):
    def finalize(self, system):
        for n, trans in enumerate(system.template_transformations):
            trans._id = n + 1

    def dump(self, system, writer):
        with writer.loop(
                "_ma_template_trans_matrix",
                ["id",
                 "rot_matrix[1][1]", "rot_matrix[2][1]", "rot_matrix[3][1]",
                 "rot_matrix[1][2]", "rot_matrix[2][2]", "rot_matrix[3][2]",
                 "rot_matrix[1][3]", "rot_matrix[2][3]", "rot_matrix[3][3]",
                 "tr_vector[1]", "tr_vector[2]", "tr_vector[3]"]) as lp:
            for t in system.template_transformations:
                lp.write(id=t._id,
                         **_get_transform(t.rot_matrix, t.tr_vector))


class _AlignmentDumper(Dumper):
    def finalize(self, system):
        for n, tmpl in enumerate(system.templates):
            tmpl._id = n + 1
        for n, segment in enumerate(system.template_segments):
            # Cannot use _id since segment might also be a complete template
            # (with _id = template id)
            segment._segment_id = n + 1
        for n, aln in enumerate(system.alignments):
            aln._id = n + 1

    def dump(self, system, writer):
        self.dump_template_details(system, writer)
        self.dump_template_poly(system, writer)
        self.dump_template_poly_segment(system, writer)
        self.dump_template_non_poly(system, writer)
        self.dump_template_ref_db(system, writer)
        self.dump_target_template_poly_mapping(system, writer)
        self.dump_template_customized(system, writer)
        self.dump_template_coord(system, writer)
        self.dump_info(system, writer)
        self.dump_details(system, writer)
        self.dump_sequences(system, writer)

    def dump_template_details(self, system, writer):
        ordinal = itertools.count(1)

        def write_template(tmpl, tgt_asym, lp):
            org = ("customized" if isinstance(tmpl, modelcif.CustomTemplate)
                   else "reference database")
            poly = ("polymer" if tmpl.entity.is_polymeric()
                    else "non-polymer")
            lp.write(ordinal_id=next(ordinal),
                     template_id=tmpl._id,
                     template_origin=org,
                     template_entity_type=poly,
                     template_trans_matrix_id=tmpl.transformation._id,
                     template_data_id=tmpl._data_id,
                     target_asym_id=tgt_asym._id if tgt_asym else None,
                     template_label_asym_id=tmpl.asym_id,
                     template_label_entity_id=tmpl.entity_id,
                     template_model_num=tmpl.model_num,
                     template_auth_asym_id=tmpl.strand_id)

        with writer.loop(
                "_ma_template_details",
                ["ordinal_id", "template_id", "template_origin",
                 "template_entity_type", "template_trans_matrix_id",
                 "template_data_id", "target_asym_id",
                 "template_label_asym_id",
                 "template_label_entity_id", "template_model_num",
                 "template_auth_asym_id"]) as lp:
            seen_templates = set()
            for a in system.alignments:
                for s in a.pairs:
                    # get Template from TemplateSegment
                    write_template(s.template.template, s.target.asym, lp)
                    seen_templates.add(s.template.template)
            # Handle all non-polymer templates (not in alignments)
            for a in system.asym_units:
                if isinstance(a, modelcif.NonPolymerFromTemplate):
                    write_template(a.template, a, lp)
                    seen_templates.add(a.template)
            # Handle all remaining non-aligned templates
            for t in system.templates:
                if t not in seen_templates:
                    write_template(t, None, lp)

    def _get_sequence(self, entity):
        """Get the sequence for an entity as a string"""
        # Split into lines to get tidier CIF output
        return "\n".join(_prettyprint_seq((comp.code if len(comp.code) == 1
                                           else '(%s)' % comp.code
                                           for comp in entity.sequence), 70))

    def _get_canon(self, entity):
        """Get the canonical sequence for an entity as a string"""
        # Split into lines to get tidier CIF output
        seq = "\n".join(_prettyprint_seq(
            (comp.code_canonical for comp in entity.sequence), 70))
        return seq

    def dump_template_poly(self, system, writer):
        with writer.loop(
                "_ma_template_poly",
                ["template_id", "seq_one_letter_code",
                 "seq_one_letter_code_can"]) as lp:
            for tmpl in system.templates:
                entity = tmpl.entity
                if not entity.is_polymeric():
                    continue
                lp.write(template_id=tmpl._id,
                         seq_one_letter_code=self._get_sequence(entity),
                         seq_one_letter_code_can=self._get_canon(entity))

    def dump_template_poly_segment(self, system, writer):
        with writer.loop("_ma_template_poly_segment",
                         ["id", "template_id", "residue_number_begin",
                          "residue_number_end"]) as lp:
            for s in system.template_segments:
                lp.write(
                    id=s._segment_id, template_id=s.template._id,
                    residue_number_begin=s.seq_id_range[0],
                    residue_number_end=s.seq_id_range[1])

    def dump_template_non_poly(self, system, writer):
        with writer.loop(
                "_ma_template_non_poly",
                ["template_id", "comp_id", "details"]) as lp:
            for tmpl in system.templates:
                entity = tmpl.entity
                if entity.is_polymeric():
                    continue
                lp.write(template_id=tmpl._id, comp_id=entity.sequence[0].id,
                         details=entity.description)

    def dump_template_ref_db(self, system, writer):
        with writer.loop(
                "_ma_template_ref_db_details",
                ["template_id", "db_name", "db_name_other_details",
                 "db_accession_code", "db_version_date"]) as lp:
            for tmpl in system.templates:
                if not isinstance(tmpl, modelcif.Template):
                    continue
                for ref in tmpl.references:
                    lp.write(template_id=tmpl._id, db_name=ref.name,
                             db_name_other_details=ref.other_details,
                             db_accession_code=ref.accession,
                             db_version_date=date.isoformat(
                                 ref.db_version_date)
                             if ref.db_version_date else None)

    def dump_template_customized(self, system, writer):
        with writer.loop(
                "_ma_template_customized", ["template_id", "details"]) as lp:
            for tmpl in system.templates:
                if isinstance(tmpl, modelcif.CustomTemplate):
                    lp.write(template_id=tmpl._id, details=tmpl.details)

    def dump_template_coord(self, system, writer):
        ordinal = itertools.count(1)
        with writer.loop(
                "_ma_template_coord",
                ["template_id", "group_PDB", "ordinal_id", "type_symbol",
                 "label_atom_id", "label_comp_id", "label_seq_id",
                 "label_asym_id", "auth_seq_id", "auth_asym_id",
                 "auth_atom_id", "auth_comp_id",
                 "Cartn_x", "Cartn_y", "Cartn_z",
                 "occupancy", "label_entity_id", "B_iso_or_equiv",
                 "formal_charge"]) as lp:
            for tmpl in system.templates:
                if not isinstance(tmpl, modelcif.CustomTemplate):
                    continue
                e = tmpl.entity
                for atom in tmpl.atoms:
                    lp.write(template_id=tmpl._id,
                             group_PDB='HETATM' if atom.het else 'ATOM',
                             ordinal_id=next(ordinal),
                             type_symbol=atom.type_symbol,
                             label_atom_id=atom.atom_id,
                             label_comp_id=e.sequence[atom.seq_id - 1].id,
                             label_seq_id=atom.seq_id,
                             label_asym_id=tmpl.asym_id,
                             auth_seq_id=atom.auth_seq_id,
                             auth_asym_id=tmpl.strand_id,
                             auth_atom_id=atom.auth_atom_id,
                             auth_comp_id=atom.auth_comp_id,
                             Cartn_x=atom.x, Cartn_y=atom.y, Cartn_z=atom.z,
                             occupancy=atom.occupancy,
                             label_entity_id=tmpl.entity_id,
                             B_iso_or_equiv=atom.biso,
                             formal_charge=atom.charge)

    def dump_target_template_poly_mapping(self, system, writer):
        ordinal = itertools.count(1)
        with writer.loop("_ma_target_template_poly_mapping",
                         ["id", "template_segment_id", "target_asym_id",
                          "target_seq_id_begin", "target_seq_id_end"]) as lp:
            for a in system.alignments:
                for p in a.pairs:
                    lp.write(
                        id=next(ordinal),
                        template_segment_id=p.template._segment_id,
                        target_asym_id=p.target.asym._id,
                        target_seq_id_begin=p.target.seq_id_range[0],
                        target_seq_id_end=p.target.seq_id_range[1])

    def dump_info(self, system, writer):
        with writer.loop(
                "_ma_alignment_info",
                ["alignment_id", "data_id", "software_group_id",
                 "alignment_length", "alignment_type",
                 "alignment_mode"]) as lp:
            for a in system.alignments:
                if a.pairs:
                    align_len = max(len(s.gapped_sequence) for pair in a.pairs
                                    for s in (pair.template, pair.target))
                else:
                    align_len = None
                lp.write(alignment_id=a._id, data_id=a._data_id,
                         software_group_id=a.software._group_id if a.software
                         else None,
                         alignment_type=a.type, alignment_mode=a.mode,
                         alignment_length=align_len,
                         alignment_type_other_details=a.other_details)

    def dump_details(self, system, writer):
        ordinal = itertools.count(1)
        with writer.loop(
                "_ma_alignment_details",
                ["ordinal_id", "alignment_id", "template_segment_id",
                 "target_asym_id", "score_type",
                 "score_type_other_details", "score_value",
                 "percent_sequence_identity",
                 "sequence_identity_denominator",
                 "sequence_identity_denominator_other_details"]) as lp:
            for a in system.alignments:
                for s in a.pairs:
                    if s.identity is None:
                        denom = od = identity = None
                    else:
                        denom = s.identity.denominator
                        od = s.identity.other_details
                        identity = s.identity.value
                    if s.score is None:
                        score_type = score_other_details = score_value = None
                    else:
                        score_type = s.score.type
                        score_other_details = s.score.other_details
                        score_value = s.score.value
                    lp.write(ordinal_id=next(ordinal), alignment_id=a._id,
                             template_segment_id=s.template._segment_id,
                             target_asym_id=s.target.asym._id,
                             score_type=score_type,
                             score_type_other_details=score_other_details,
                             score_value=score_value,
                             percent_sequence_identity=identity,
                             sequence_identity_denominator=denom,
                             sequence_identity_denominator_other_details=od)

    def dump_sequences(self, system, writer):
        ordinal = itertools.count(1)
        with writer.loop(
                "_ma_alignment",
                ["ordinal_id", "alignment_id", "target_template_flag",
                 "sequence"]) as lp:
            for a in system.alignments:
                # todo: don't duplicate sequences
                for s in a.pairs:
                    # 1=target, 2=template
                    lp.write(ordinal_id=next(ordinal), alignment_id=a._id,
                             target_template_flag=1,
                             sequence=s.target.gapped_sequence)
                    lp.write(ordinal_id=next(ordinal), alignment_id=a._id,
                             target_template_flag=2,
                             sequence=s.template.gapped_sequence)


class _ProtocolDumper(Dumper):
    def finalize(self, system):
        # Assign IDs to protocols and steps
        for np, p in enumerate(system.protocols):
            p._id = np + 1
            for ns, s in enumerate(p.steps):
                s._id = ns + 1

    def dump(self, system, writer):
        ordinal = itertools.count(1)
        with writer.loop(
                "_ma_protocol_step",
                ['ordinal_id', 'protocol_id', 'step_id', 'method_type',
                 'step_name', 'details', 'software_group_id',
                 'input_data_group_id', 'output_data_group_id']) as lp:
            for p in system.protocols:
                for s in p.steps:
                    lp.write(ordinal_id=next(ordinal), protocol_id=p._id,
                             step_id=s._id, method_type=s.method_type,
                             step_name=s.name, details=s.details,
                             software_group_id=s.software._group_id
                             if s.software else None,
                             input_data_group_id=s.input_data._data_group_id
                             if s.input_data else None,
                             output_data_group_id=s.output_data._data_group_id
                             if s.output_data else None)


class _ModelDumper(ihm.dumper._ModelDumperBase):
    def dump(self, system, writer):
        self.dump_model_list(system, writer)
        self.dump_model_groups(system, writer)
        seen_types = self.dump_atoms(system, writer, add_ihm=False)
        self.dump_atom_type(seen_types, system, writer)

    def dump_model_list(self, system, writer):
        with writer.loop("_ma_model_list",
                         ["ordinal_id", "model_name",
                          "data_id", "model_type",
                          "model_type_other_details"]) as lp:
            for group, model in sorted(system._all_models(),
                                       key=lambda x: x[1]._id):
                lp.write(ordinal_id=model._id, model_name=model.name,
                         data_id=model._data_id, model_type=model.model_type,
                         model_type_other_details=model.other_details)

    def dump_model_groups(self, system, writer):
        self.dump_model_group_summary(system, writer)
        self.dump_model_group_link(system, writer)

    def dump_model_group_summary(self, system, writer):
        with writer.loop("_ma_model_group", ["id", "name", "details"]) as lp:
            for group in system.model_groups:
                # ihm.model.ModelGroup only supports details after v1.8
                lp.write(id=group._id, name=group.name,
                         details=group.details
                         if hasattr(group, 'details') else None)

    def dump_model_group_link(self, system, writer):
        with writer.loop("_ma_model_group_link",
                         ["group_id", "model_id"]) as lp:
            for group in system.model_groups:
                for model_id in sorted(set(model._id for model in group)):
                    lp.write(model_id=model_id, group_id=group._id)


class _AssociatedDumper(Dumper):
    def finalize(self, system):
        file_id = itertools.count(1)
        in_archive_file_id = itertools.count(1)
        for repo in system.repositories:
            for f in repo.files:
                f._id = next(file_id)
                if hasattr(f, 'files'):
                    for af in f.files:
                        if hasattr(af, 'files'):
                            raise ValueError(
                                "An archive cannot contain another archive")
                        af._id = next(in_archive_file_id)

    def dump(self, system, writer):
        self.dump_files(system, writer)
        self.dump_archive_files(system, writer)

    def dump_files(self, system, writer):
        with writer.loop(
                "_ma_entry_associated_files",
                ["id", "entry_id", "file_url", "file_type", "file_format",
                 "file_content", "details", "data_id"]) as lp:
            for repo in system.repositories:
                for f in repo.files:
                    lp.write(id=f._id, entry_id=system.id,
                             file_url=repo.get_url(f), file_type=f.file_type,
                             file_format=f.file_format,
                             file_content=f.file_content, details=f.details,
                             data_id=f.data._data_id if f.data else None)

    def dump_archive_files(self, system, writer):
        with writer.loop(
                "_ma_associated_archive_file_details",
                ["id", "archive_file_id", "file_path", "file_format",
                 "file_content", "description", "data_id"]) as lp:
            for repo in system.repositories:
                for f in repo.files:
                    if not hasattr(f, 'files'):
                        continue
                    for af in f.files:
                        lp.write(id=af._id, archive_file_id=f._id,
                                 file_path=af.path, file_format=af.file_format,
                                 file_content=af.file_content,
                                 description=af.details,
                                 data_id=af.data._data_id if af.data else None)


class _FeatureDumper(Dumper):
    def finalize(self, system):
        seen_features = {}
        self._features_by_id = []
        for f in system._all_features():
            util._remove_id(f)
        for f in system._all_features():
            util._assign_id(f, seen_features, self._features_by_id,
                            seen_obj=f._signature())

    def dump(self, system, writer):
        self.dump_list(writer)
        self.dump_atom(writer)
        self.dump_residue(writer)
        self.dump_instance(writer)

    def dump_list(self, writer):
        with writer.loop("_ma_feature_list",
                         ["feature_id", "feature_type", "entity_type",
                          "details"]) as lp:
            for f in self._features_by_id:
                lp.write(feature_id=f._id, feature_type=f.type,
                         entity_type=f._get_entity_type(check=self._check),
                         details=f.details)

    def dump_atom(self, writer):
        ordinal = itertools.count(1)
        with writer.loop("_ma_atom_feature",
                         ["ordinal_id", "feature_id", "atom_id"]) as lp:
            for f in self._features_by_id:
                if not isinstance(f, modelcif.AtomFeature):
                    continue
                for a in f.atoms:
                    lp.write(ordinal_id=next(ordinal), feature_id=f._id,
                             atom_id=a)

    def dump_residue(self, writer):
        ordinal = itertools.count(1)
        with writer.loop("_ma_poly_residue_feature",
                         ["ordinal_id", "feature_id", "label_asym_id",
                          "label_seq_id", "label_comp_id"]) as lp:
            for f in self._features_by_id:
                if not isinstance(f, modelcif.PolyResidueFeature):
                    continue
                for r in f.residues:
                    seq = r.entity.sequence
                    lp.write(ordinal_id=next(ordinal), feature_id=f._id,
                             label_asym_id=r.asym._id,
                             label_seq_id=r.seq_id,
                             label_comp_id=seq[r.seq_id - 1].id)

    def dump_instance(self, writer):
        ordinal = itertools.count(1)
        with writer.loop("_ma_entity_instance_feature",
                         ["ordinal_id", "feature_id", "label_asym_id"]) as lp:
            for f in self._features_by_id:
                if not isinstance(f, modelcif.EntityInstanceFeature):
                    continue
                for a in f.asym_units:
                    lp.write(ordinal_id=next(ordinal), feature_id=f._id,
                             label_asym_id=a._id)


class _QAMetricDumper(Dumper):
    def finalize(self, system):
        # Get all metric classes used by all systems
        seen_metric_classes = set()
        self._metric_classes_by_id = []
        metric_id = itertools.count(1)
        for group, model in system._all_models():
            for m in model.qa_metrics:
                cls = type(m)
                if cls not in seen_metric_classes:
                    seen_metric_classes.add(cls)
                    cls._id = next(metric_id)
                    # We need an instance of the class in case name or
                    # description are provided by property()
                    self._metric_classes_by_id.append(m)

    def dump(self, system, writer):
        self.dump_metric_types(system, writer)
        self.dump_metric_global(system, writer)
        self.dump_metric_local(system, writer)
        self.dump_metric_pairwise(system, writer)
        self.dump_metric_feature(system, writer)
        self.dump_metric_feature_pairwise(system, writer)
        self.dump_metric_dihedral(system, writer)

    def dump_metric_types(self, system, writer):
        with writer.loop(
                "_ma_qa_metric",
                ["id", "name", "description", "type", "mode",
                 "type_other_details", "software_group_id"]) as lp:
            for m in self._metric_classes_by_id:
                lp.write(id=m._id, name=m.name, description=m.description,
                         type=m.type, mode=m.mode,
                         type_other_details=m.other_details,
                         software_group_id=m.software._group_id if m.software
                         else None)

    def dump_metric_global(self, system, writer):
        ordinal = itertools.count(1)
        with writer.loop(
                "_ma_qa_metric_global",
                ["ordinal_id", "model_id", "metric_id", "metric_value"]) as lp:
            for group, model in system._all_models():
                for m in model.qa_metrics:
                    if not isinstance(m, modelcif.qa_metric.Global):
                        continue
                    lp.write(ordinal_id=next(ordinal), model_id=model._id,
                             metric_id=m._id, metric_value=m.value)

    def dump_metric_local(self, system, writer):
        ordinal = itertools.count(1)
        with writer.loop(
                "_ma_qa_metric_local",
                ["ordinal_id", "model_id", "label_asym_id", "label_seq_id",
                 "label_comp_id", "metric_id", "metric_value"]) as lp:
            for group, model in system._all_models():
                for m in model.qa_metrics:
                    if not isinstance(m, modelcif.qa_metric.Local):
                        continue
                    seq = m.residue.asym.entity.sequence
                    lp.write(ordinal_id=next(ordinal), model_id=model._id,
                             label_asym_id=m.residue.asym._id,
                             label_seq_id=m.residue.seq_id,
                             label_comp_id=seq[m.residue.seq_id - 1].id,
                             metric_id=m._id, metric_value=m.value)

    def dump_metric_pairwise(self, system, writer):
        ordinal = itertools.count(1)
        with writer.loop(
                "_ma_qa_metric_local_pairwise",
                ["ordinal_id", "model_id", "label_asym_id_1", "label_seq_id_1",
                 "label_comp_id_1", "label_asym_id_2", "label_seq_id_2",
                 "label_comp_id_2", "metric_id", "metric_value"]) as lp:
            for group, model in system._all_models():
                for m in model.qa_metrics:
                    if not isinstance(m, modelcif.qa_metric.LocalPairwise):
                        continue
                    seq1 = m.residue1.asym.entity.sequence
                    seq2 = m.residue2.asym.entity.sequence
                    lp.write(ordinal_id=next(ordinal), model_id=model._id,
                             label_asym_id_1=m.residue1.asym._id,
                             label_seq_id_1=m.residue1.seq_id,
                             label_comp_id_1=seq1[m.residue1.seq_id - 1].id,
                             label_asym_id_2=m.residue2.asym._id,
                             label_seq_id_2=m.residue2.seq_id,
                             label_comp_id_2=seq2[m.residue2.seq_id - 1].id,
                             metric_id=m._id, metric_value=m.value)

    def dump_metric_feature(self, system, writer):
        ordinal = itertools.count(1)
        with writer.loop(
                "_ma_qa_metric_feature",
                ["ordinal_id", "model_id", "feature_id", "metric_id",
                 "metric_value"]) as lp:
            for group, model in system._all_models():
                for m in model.qa_metrics:
                    if not isinstance(m, modelcif.qa_metric.Feature):
                        continue
                    lp.write(ordinal_id=next(ordinal), model_id=model._id,
                             feature_id=m.feature._id,
                             metric_id=m._id, metric_value=m.value)

    def dump_metric_feature_pairwise(self, system, writer):
        ordinal = itertools.count(1)
        with writer.loop(
                "_ma_qa_metric_feature_pairwise",
                ["ordinal_id", "model_id", "feature_id_1", "feature_id_2",
                 "metric_id", "metric_value"]) as lp:
            for group, model in system._all_models():
                for m in model.qa_metrics:
                    if not isinstance(m, modelcif.qa_metric.FeaturePairwise):
                        continue
                    lp.write(ordinal_id=next(ordinal), model_id=model._id,
                             feature_id_1=m.feature1._id,
                             feature_id_2=m.feature2._id,
                             metric_id=m._id, metric_value=m.value)

    def dump_metric_dihedral(self, system, writer):
        ordinal = itertools.count(1)
        with writer.loop(
                "_ma_qa_metric_dihedral",
                ["ordinal_id", "atom_id_1", "atom_id_2", "atom_id_3",
                 "atom_id_4", "metric_id", "metric_value", "quality",
                 "smarts_pattern"]) as lp:
            for group, model in system._all_models():
                for m in model.qa_metrics:
                    if not isinstance(m, modelcif.qa_metric.Dihedral):
                        continue
                    lp.write(ordinal_id=next(ordinal), atom_id_1=m.atom_id_1,
                             atom_id_2=m.atom_id_2, atom_id_3=m.atom_id_3,
                             atom_id_4=m.atom_id_4, metric_id=m._id,
                             metric_value=m.value, quality=m.quality,
                             smarts_pattern=m.smarts_pattern)


class _CopyWriter:
    """Context manager to write loop or category to two mmCIF/BinaryCIF
       files"""
    def __init__(self, w1, w2):
        self.w1, self.w2 = w1, w2

    def write(self, *args, **keys):
        self.w1.write(*args, **keys)
        self.w2.write(*args, **keys)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # This may not correctly handle exceptions raised within the loop
        self.w1.__exit__(exc_type, exc_value, traceback)
        self.w2.__exit__(exc_type, exc_value, traceback)


class _SystemWriter:
    """Utility class which normally just passes through to the default
       ``base_writer``, but outputs selected categories to associated files."""
    def __init__(self, base_writer, category_map, copy_category_map):
        self._base_writer = base_writer
        self.category_map = category_map
        self.copy_category_map = copy_category_map

    def category(self, category):
        w = self.copy_category_map.get(category)
        if w:
            return _CopyWriter(w.category(category),
                               self._base_writer.category(category))
        else:
            w = self.category_map.get(category, self._base_writer)
            return w.category(category)

    def loop(self, category, keys):
        w = self.copy_category_map.get(category)
        if w:
            return _CopyWriter(w.loop(category, keys),
                               self._base_writer.loop(category, keys))
        else:
            w = self.category_map.get(category, self._base_writer)
            return w.loop(category, keys)

    def end_block(self):
        # Flush and close all file handles of associated files
        for w in self.category_map.values():
            if not hasattr(w, 'fh'):
                continue
            w.flush()
            w.fh.close()
            del w.fh

    # Just pass through to base writer object
    def flush(self):
        return self._base_writer.flush()

    def start_block(self, name):
        return self._base_writer.start_block(name)

    def write_comment(self, comment):
        return self._base_writer.write_comment(comment)


class ModelCIFVariant(Variant):
    """Used to select typical PDBx/ModelCIF file output.
       See :func:`write` and :class:`ihm.dumper.Variant`."""
    _dumpers = [
        ihm.dumper._EntryDumper,  # must be first
        ihm.dumper._StructDumper, ihm.dumper._CommentDumper,
        _AuditConformDumper, _DatabaseDumper, ihm.dumper._CitationDumper,
        ihm.dumper._SoftwareDumper, _SoftwareGroupDumper,
        ihm.dumper._AuditAuthorDumper, ihm.dumper._AuditRevisionDumper,
        ihm.dumper._DataUsageDumper, ihm.dumper._GrantDumper,
        _ChemCompDumper, _ChemCompDescriptorDumper,
        ihm.dumper._EntityDumper,
        ihm.dumper._EntitySrcGenDumper, ihm.dumper._EntitySrcNatDumper,
        ihm.dumper._EntitySrcSynDumper, ihm.dumper._StructRefDumper,
        _TargetRefDBDumper,
        ihm.dumper._EntityPolyDumper, _EntityNonPolyDumper,
        ihm.dumper._EntityPolySeqDumper, ihm.dumper._EntityBranchListDumper,
        ihm.dumper._EntityBranchDumper, ihm.dumper._StructAsymDumper,
        ihm.dumper._PolySeqSchemeDumper, ihm.dumper._NonPolySchemeDumper,
        ihm.dumper._BranchSchemeDumper, ihm.dumper._BranchDescriptorDumper,
        ihm.dumper._BranchLinkDumper,
        _DataDumper, _DataGroupDumper, _DataRefDBDumper,
        _TargetEntityDumper, _TemplateTransformDumper, _AlignmentDumper,
        _ProtocolDumper, _ModelDumper, _AssociatedDumper, _FeatureDumper,
        _QAMetricDumper]

    def get_dumpers(self):
        return [d() for d in self._dumpers]

    def get_system_writer(self, system, writer_class, writer):
        # Get a Writer-like object which outputs selected categories to
        # associated files (the rest use the default writer)
        category_map = {}
        copy_category_map = {}

        def _all_repo_files(r):
            for f in r.files:
                yield f
                if hasattr(f, 'files'):
                    for subf in f.files:
                        yield subf
        for r in system.repositories:
            for f in _all_repo_files(r):
                if (not hasattr(f, 'categories')
                        or (not f.categories and not f.copy_categories)):
                    continue
                if f.binary:
                    w = ihm.format_bcif.BinaryCifWriter(
                        open(f.local_path, 'wb'))
                else:
                    w = ihm.format.CifWriter(open(f.local_path, 'w'))
                # Write header information to the associated file
                dumpers = (ihm.dumper._EntryDumper(), _EntryLinkDumper())
                # We are passing the File object to the dumpers here where
                # they expect a System object, but the interfaces are similar
                # enough, so we don't need a facade object.
                for d in dumpers:
                    d.finalize(f)
                for d in dumpers:
                    d.dump(f, w)
                for c in f.categories:
                    # Allow for categories with or without leading underscore
                    category_map['_' + c.lstrip('_').lower()] = w
                for c in f.copy_categories:
                    copy_category_map['_' + c.lstrip('_').lower()] = w
        if category_map or copy_category_map:
            return _SystemWriter(writer, category_map, copy_category_map)
        else:
            # If no categories, we can just use the base writer
            return writer


def write(fh, systems, format='mmCIF', dumpers=[],
          variant=ModelCIFVariant, check=True):
    """Write out all `systems` to the file handle `fh`.

       See :func:`ihm.dumper.write` for more information. The function
       here behaves similarly but writes out files compliant with the
       ModelCIF extension directory rather than IHM."""
    return ihm.dumper.write(fh, systems, format, dumpers, variant, check=check)



"""Classes for handling modeling protocols.
"""


class Step:
    """A single step in a :class:`Protocol`.

       This class describes a generic step in a modeling protocol. In most
       cases, a more specific subclass should be used, such as
       :class:`TemplateSearchStep`, :class:`ModelingStep`, or
       :class:`ModelSelectionStep`.

       :param input_data: Any objects that this step takes as input.
              Any individual :class:`modelcif.data.Data` object (such as
              a template structure, target sequence, alignment, or model
              coordinates) can be given here, or a group of such objects (as a
              :class:`modelcif.data.DataGroup` object) can be passed.
       :type input_data: :class:`modelcif.data.DataGroup`
             or :class:`modelcif.data.Data`
       :param output_data: Any objects that this step creates as output,
              similarly to ``input_data``.
       :type output_data: :class:`modelcif.data.DataGroup`
             or :class:`modelcif.data.Data`
       :param str name: A short name for this step.
       :param str details: Longer description of this step.
       :param software: The software that was employed in this modeling step.
       :type software: :class:`modelcif.Software`
             or :class:`modelcif.SoftwareGroup`
    """
    method_type = "other"

    def __init__(self, input_data, output_data, name=None, details=None,
                 software=None):
        self.input_data, self.output_data = input_data, output_data
        self.name, self.details, self.software = name, details, software


class TemplateSearchStep(Step):
    """A modeling protocol step that searches for templates.
       See :class:`Step` for more details."""
    method_type = "template search"


class TemplateSelectionStep(Step):
    """A modeling protocol step that selects one or more templates.
       See :class:`Step` for more details."""
    method_type = "template selection"


class TargetTemplateAlignmentStep(Step):
    """A modeling protocol step that creates a target-template alignment.
       See :class:`Step` for more details."""
    method_type = "target-template alignment"


class CoevolutionMSAStep(Step):
    """A modeling protocol step that creates a coevolution multiple sequence
       alignment. See :class:`Step` for more details."""
    method_type = "coevolution MSA"


class ContactPredictionStep(Step):
    """A modeling protocol step that predicts contacts.
       See :class:`Step` for more details."""
    method_type = "contact prediction"


class ModelingStep(Step):
    """A modeling protocol step that generates model coordinates.
       See :class:`Step` for more details."""
    method_type = "modeling"


class ModelSelectionStep(Step):
    """A modeling protocol step that filters candidates to select models.
       See :class:`Step` for more details."""
    method_type = "model selection"


class ModelRefinementStep(Step):
    """A modeling protocol step that refines models.
       See :class:`Step` for more details."""
    method_type = "model refinement"


class ModelAssessmentStep(Step):
    """A modeling protocol step that performs model quality assessment.
       See :class:`Step` for more details."""
    method_type = "model quality assessment"


class BackboneGenerationStep(Step):
    """A modeling protocol step that generates backbone coordinates for
       protein design.
       See :class:`Step` for more details."""
    method_type = "backbone generation"


class SequenceDesignStep(Step):
    """A modeling protocol step that generates sequences for protein design.
       This includes inverse folding.
       See :class:`Step` for more details."""
    method_type = "sequence design"


class StructureValidationStep(Step):
    """A modeling protocol step that assesses designed sequences with
       their modeled structures.
       See :class:`Step` for more details."""
    method_type = "structure validation"


class EnergyEstimateStep(Step):
    """A modeling protocol step that performs an energy estimate.
       See :class:`Step` for more details."""
    method_type = "energy estimate"


class ExperimentalValidationStep(Step):
    """A modeling protocol step that conducts experimental validation.
       See :class:`Step` for more details."""
    method_type = "experimental validation"


class Protocol:
    """A modeling protocol.
       Each protocol consists of a number of protocol steps."""
    def __init__(self):
        #: All modeling steps (:class:`Step` objects)
        self.steps = []



import utils
import os
import unittest

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import modelcif.alignment


class Tests(unittest.TestCase):
    def test_identity(self):
        """Test sequence identity classes"""
        ident = modelcif.alignment.ShorterSequenceIdentity(42.0)
        self.assertEqual(ident.denominator, "Length of the shorter sequence")
        self.assertIsNone(ident.other_details)
        self.assertAlmostEqual(ident.value, 42.0, delta=1e-4)
        ident = modelcif.alignment.AlignedPositionsIdentity(42.0)
        ident = modelcif.alignment.AlignedResiduePairsIdentity(42.0)
        ident = modelcif.alignment.MeanSequenceIdentity(42.0)

        # generic "other" identity
        ident = modelcif.alignment.Identity(42.0)
        self.assertEqual(ident.denominator, "Other")
        self.assertIsNone(ident.other_details)

        # custom "other" identity
        class CustomIdentity(modelcif.alignment.Identity):
            """foo
               bar"""

        ident = CustomIdentity(42.0)
        self.assertEqual(ident.denominator, "Other")
        self.assertEqual(ident.other_details, "foo")


if __name__ == '__main__':
    unittest.main()



import os
import unittest
import utils

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import modelcif.qa_metric


class Tests(unittest.TestCase):
    def test_metric_types(self):
        """Test MetricType subclasses"""
        self.assertEqual(modelcif.qa_metric.Energy.type, "energy")
        self.assertIsNone(modelcif.qa_metric.Energy.other_details)
        self.assertEqual(modelcif.qa_metric.PAE.type, "PAE")
        self.assertEqual(modelcif.qa_metric.ContactProbability.type,
                         "contact probability")

        # MetricType itself should have no other_details
        class Custom1(modelcif.qa_metric.Global,
                      modelcif.qa_metric.MetricType):
            """Custom 1"""
        x = Custom1(42)
        self.assertEqual(x.type, "other")
        self.assertIsNone(x.other_details)

        class CustomMetricType(modelcif.qa_metric.MetricType):
            """foo
               bar"""

        class Custom2(modelcif.qa_metric.Global, CustomMetricType):
            """Custom 2"""
        x = Custom2(42)
        self.assertEqual(x.type, "other")
        self.assertEqual(x.other_details, "foo")

        # MetricType in the enumeration should have no "other_details"

        class EnumMetricType(modelcif.qa_metric.MetricType):
            """foo"""
            type = "enum"

        class Custom3(modelcif.qa_metric.Global, EnumMetricType):
            """Custom 3"""
        x = Custom3(42)
        self.assertEqual(x.type, "enum")
        self.assertIsNone(x.other_details)

    def test_global_metric(self):
        """Test Global MetricMode"""
        class MyScore(modelcif.qa_metric.Global, modelcif.qa_metric.Energy):
            pass

        q = MyScore(42)
        _ = repr(q)

    def test_local_metric(self):
        """Test Local MetricMode"""
        class MyScore(modelcif.qa_metric.Local, modelcif.qa_metric.Energy):
            pass

        e1 = modelcif.Entity('ACGT')
        asym = modelcif.AsymUnit(e1, 'foo')
        q = MyScore(asym.residue(2), 42)
        _ = repr(q)

    def test_local_pairwise_metric(self):
        """Test LocalPairwise MetricMode"""
        class MyScore(modelcif.qa_metric.LocalPairwise,
                      modelcif.qa_metric.Energy):
            pass

        e1 = modelcif.Entity('ACGT')
        asym = modelcif.AsymUnit(e1, 'foo')
        q = MyScore(asym.residue(2), asym.residue(3), 42)
        _ = repr(q)

    def test_feature_metric(self):
        """Test Feature MetricMode"""
        class MyScore(modelcif.qa_metric.Feature, modelcif.qa_metric.Energy):
            pass

        e1 = modelcif.Entity('ACGT')
        asym = modelcif.AsymUnit(e1, 'foo')
        q = MyScore(asym.residue(2), 42)
        _ = repr(q)

    def test_feature_pairwise_metric(self):
        """Test FeaturePairwise MetricMode"""
        class MyScore(modelcif.qa_metric.FeaturePairwise,
                      modelcif.qa_metric.Energy):
            pass

        e1 = modelcif.Entity('ACGT')
        asym = modelcif.AsymUnit(e1, 'foo')
        q = MyScore(asym.residue(2), asym.residue(3), 42)
        _ = repr(q)

    def test_dihedral_metric(self):
        """Test Dihedral MetricMode"""
        class MyScore(modelcif.qa_metric.Dihedral,
                      modelcif.qa_metric.Energy):
            pass

        q = MyScore(1, 2, 3, 4, 42, 'relaxed')
        _ = repr(q)


if __name__ == '__main__':
    unittest.main()
